# """A set of classes to handle the NaCl training datasets

# The NaCl models are trained on lightcurves and spectra. The TrainingDataset
# objects presents the models with a uniform interface, which allows to
# navigate easily in this hybrid dataset. TrainingDataset objects have three
# main purposes: (1) they maintain the relations between the SN meta
# information (z, identifier, x0, ...) and the related light curves and spectra
# (2) they map the SN, lightcurve or spectra unique ID's (int or str) into
# continuous integer indexes (3) they allow to track whether a data point is
# valid or not (i.e.  whether it should be included in the likelihood
# computation or not) (4) they maintain access to the passband shapes.

# Tests have shown that accessing data stored in a pandas DataFrame is
# significantly slower than slice-based access on raw numpy arrays (by about a
# factor 10 as of writing this code). For this reason, we have decided to keep
# our old implementation and store the data internally into `numpy.recarray`s.
# A TrainingDataset may be passed numpy arrays or pandas DataFrame. The data is
# sorted to ensure that light curves and spectra of a same SN are stored
# contiguously in memory.

# Internals, format
# -----------------

# A TrainingDataset maintains 3 main arrays, encapsulated into
# `nacl.lib.dataproxy.DataProxy`'s objects. The dataproxies maintain the
# contiguous indexing described above, and allow to seamlessly add fields to
# the arrays.

# A) `TrainingDataset` needs 3 inputs: lc_data, spec_data, sne. All the
# elements are dataframes or recarrays

#  `sn_data`: Information regarding the SN. Need at 8 columns.
#   One row corresponds to one SN. The minimal amount of information
#   required is:
#      - sn : a unique identifier (int or str) for the SN.
#        May be the TNS/IAU name
#      - z : heliocentric redshfit
#      - valid : whether the SN (and all its follow-up data) is valid
#      - tmax : time of the maximum in B band
#      - x1 : x1 parameter from SALT2
#      - x0 : x0 parameter from SALT2
#      - c : color parameter from SALT2


# Example of a sn_data row:
#     sn          z         tmax         x1         x0        col       valid
#     b'sn1981b'  0.006030  44672.524515 -0.423846  0.294523  0.042690    1


# B) `lc_data`: a larger array that contains all the photometric follow-up
# points.  Needs at least 9 columns. One row corresponds to one photometric
# point.

#     - sn : sn unique identifier, as specified in `sn_data` (see above)
#     - lc : unique light curve identifier
#     - mjd : observation date (Modified Julian Day)
#     - flux : flux
#     - fluxerr : flux error
#     - band : band name  instrument::filter, e.g STANDARD::B or LSST::g
#     - magsys : magnitude system, e.g., Vega2, AB_B12C,
#     - mag_sky : magnitude of the sky
#     - seeing : seeing of the night
#     - exptime : Exposure time. NaN when you use real data like JLA, K21
#     - valid : 0 or 1. 0 the data point is bad, we do not use it

# Example of Lc_data:
# sn   mjd        flux       fluxerr   band             magsys     exptime  valid    lc	seeing	mag_sys
# 0    44669.300  18.685584  0.913621  b'STANDARD::B'    VEGA2      NaN      1       0	NaN	NaN


# C) `spec_data`: Information regarding the spectroscopy. Need at 8 columns.
#     One row corresponds to one spectrum bin.

#     - sn : sn unique identifier, as specified in `sn_data` (see above)
#     - spec : spectrum unique identifier
#     - mjd : spectrum observation date (Modified Julian Day)
#     - wavelength : observer frame wavelength in Angs
#     - flux : flux
#     - fluxerr : flux error
#     - exptime : Exposure time (seconds) NaN when you use real data
#                 like JLA, K21
#     - valid : 0 or 1. 0 the spectrun data point is bad, we do not use it

# Example of spec_data:
# sn   mjd      wavelength   flux          fluxerr        valid   spec  exptime
# 107  52905.0  3739.254     5.588000e-17  1.792800e-17   1       0     NaN


# Indexes
# -------


# Input/Output
# ------------
# A TrainingDataset may be written on disk in several different formats:
# pickle, numpy:npz (compressed or not) and parquet.  We recommend parquet,
# which seems to offer the best performances in terms of speed and compression.

# # to create a dataset, we just have to pass the sn, lc and spec arrays
# # they can be passed as numpy.recarrays or as pandas.DataFrames
# >>> tds = TrainingDataset(lc_data, spec_data, sample)

# # a dataset can be saved to disk and re-read in several formats.
# # we recommend parquet, but npz and pickle are also available
# >>> tds.to_parquet(name='JLA', path='./')
# >>> tds = nacl.dataset.read_parquet(name='JLA', path='./')

# # it is easy to invalidate data. This functionality is used
# # for example, during outlier removal. The corresponding data entries
# # are just flagged, and subsequently ignored in the likelihood computation.
# >>> tds.kill_sne([58]) # specify the SN unique ID
# >>> tds.kill_lc([729])
# >>> tds.kill_spec([1255, 3280])

# if one want to get rid of the data for real, it is possible
# to use .compress() -- costly however, since it necessitates rebuilding
# the indexes.
# >>> tds.compress()

# .. todo :: add a piece of code in the constructor of TrainingDataset,
#            to make sure that no name collisions
#            (i.e. SN identifier is unique).
# """

import logging
import pathlib
from typing import List
from copy import copy as cp
import re
import pickle

import numpy as np
import pandas

import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt

# from nacl.handles import LcData, SNData, SpectrumData
from saltworks import DataProxy


from sncosmo._registry import Registry
import sncosmo


_TRAINING_DATASETS = Registry()


def get_training_dataset(name, version=None, copy=False):
    """Retrieve a training dataset from the registry

    Parameters
    ----------
    name : str
        Name of training dataset in the registry
    version : str, optional
        version identifier for sources with multiple versions. Default
        is `None` (i.e. latest version)
    copy : bool, optional
       If True, and if `name` already a Source instance, return a copy of it.
    """
    if isinstance(name, TrainingDataset):
        if copy:
            return cp(name)
        else:
            return name
    else:
        return cp(_TRAINING_DATASETS.retrieve(name, version=version))

class TrainingDatasetError(Exception):
    pass


def opt_band_name_order(bn):
    _, band = re.match('(.+)::([ugrizyUBVRIYJHK]+)', bn).groups()
    return 'ugrizyUBVRIYJHK'.index(band)


class TrainingDataset:
    """A class to hold and index the training dataset.

    A typical training dataset contains light curves, spectra, plus
    meta-information to link these light curves and spectra to their respective
    SNe, plus meta information relative to these SNe (e.g. redshift).

    The TrainingDataset is used to organize and index the spectroscopic and
    photometric data and present an abstract interface to the models.

    """

    sn_data_dtype = np.dtype(
        [('index', '<i8'), ('sn', '<i8'), ('z', '<f8'),
         ('tmax', '<f8'), ('x1', '<f8'), ('x0', '<f8'), ('c', '<f8'),
         ('ebv', '<f8'), ('valid', '<i8'), ('IAU', 'O')]
    )

    lc_data_dtype = np.dtype(
        [('index', '<i8'), ('sn', '<i8'),
         ('mjd', '<f8'), ('flux', '<f8'), ('fluxerr', '<f8'),
         ('band', 'O'), ('magsys', 'O'), ('exptime', '<f8'),
         ('valid', '<i8'), ('lc', '<i8'),
         ('zp', '<f8'), ('mag_sky', '<f8'), ('seeing', '<f8'), ('x', '<f8'), ('y', '<f8'), ('sensor_id', '<i8')]
    )

    # was do we need exptime here ?
    spec_data_dtype = np.dtype(
        [('index', '<i8'), ('sn', '<i8'),
         ('mjd', '<f8'), ('wavelength', '<f8'), ('flux', '<f8'), ('fluxerr', '<f8'),
         ('i_basis', '<i8'),
         ('valid', '<i8'), ('spec', '<i8'), ('exptime', '<f8')]
    )

    def __init__(self, sne, lc_data=None, spec_data=None, spectrophotometric_data=None,
                 filterpath=None, spectra=None, lcs=None, filterlib=None, basis=None):
        """load the file, sort by data type (lc/spec) and SN

        Parameters
        ----------
        sne : array-like or DataFrame
            sn catalog (sn, name, z, X0, X1, c...)
        lc_data : recarray or DataFrame
            light curve photometric data.
        spec_data : recarray or DataFrame
            spectra
        spectrophotometric_data : recarray or DataFrame
            spectrophotometric data (e.g. SNIFS spectra)
        """
        sne = _df_to_recarray(sne)
        self.sn_data = DataProxy(sne, sn='sn', z='z',
                                 tmax='tmax', x0='x0', x1='x1', col='c',
                                 valid='valid')

        # at this stage, we convert everything into recarrays in order to
        # minimize the access times. We found indeed that access times
        # to a pandas.DataFrame can be up to an order of magnitude slower
        # than access time to a numpy.recarray
        self.lc_data = None
        self.spec_data = None
        self.spectrophotometric_data = None

        if lc_data is not None:
            lc_data = _df_to_recarray(lc_data, sort_by=['sn', 'band'])
            self.lc_data = DataProxy(lc_data, sn='sn', lc='lc', mjd='mjd',
                                     band='band', exptime='exptime',
                                     magsys='magsys', zp='zp',
                                     mag_sky='mag_sky', seeing='seeing',
                                     flux='flux', fluxerr='fluxerr',
                                     x='x', y='y', sensor_id='sensor_id',
                                     valid='valid')
            self.update_zp_scale()

        if spec_data is not None:
            spec_data = _df_to_recarray(spec_data)  # , sort_by=['sn', 'mjd'])
            self.spec_data = DataProxy(spec_data, sn='sn', spec='spec',
                                       mjd='mjd',
                                       wavelength='wavelength',
                                       flux='flux', fluxerr='fluxerr',
                                       i_basis='i_basis',
                                       valid='valid')

        if spectrophotometric_data is not None:
            spectrophotometric_data = _df_to_recarray(spectrophotometric_data)
            self.spectrophotometric_data = DataProxy(spectrophotometric_data, sn='sn',
                                                     spec='spec', mjd='mjd',
                                                     wavelength='wavelength',
                                                     flux='flux',
                                                     fluxerr='fluxerr',
                                                     i_basis='i_basis',
                                                     valid='valid')

        # we need to have either of these
        # assert(self.lc_data or self.spectrophotometric_data)

        # we probably need to update this after a compress ... ?
        # load all the transmissions and a mean_wavelength field
        # to the light curve data
        self.filterlib = filterlib
        if filterlib is not None:
            if not self.lc_data:
                logging.warning('will not load filters - no photometric data in the training dataset')
            else:
                trans = self.get_all_transmissions()
                self.band_wavelength = dict([(nm, tr.wave_eff()) for nm, tr in trans.items()])
                wl = np.array([self.band_wavelength[k] for k in self.lc_data.band])
                self.lc_data.add_field('wavelength', wl)
                # self.get_magsys_zp()
                # self.compute_photometric_norm_factors()

        # build a general index for the SN data
        self._index_sne()

        # build an index for the LC data
        # and connect it to the SN index
        if lcs is None and self.lc_data:
            self.lcs = []
            self._index_lcs()
        else:
            self.lcs = lcs

        # build an index for the spectroscopic data
        # and connect it to the SN index
        self.spectra = []
        self._index_spectra()

        # if we have spectrophotometric data available
        # (e.g. the SNfactory dataset), then build an index
        # of these spectra
        if self.spectrophotometric_data is not None:
            self._index_photometric_spectra()

        # build a global index. it is costly, but it can help
        # self.sne, self.lcs, self.spectra = build_index(self)

        self.plotter = None
        
        self.basis = basis
        
    @staticmethod
    def from_hdf(path):
        """
        """
        tds = read_hdf(path)
        return tds

    def extract(self, sn):
        """extract just the sn data
        """
        idx = self.sn_data.sn == sn
        if idx.sum() == 0:
            return None
        sn_data = self.sn_data.nt[idx]

        lc_data = None
        if self.lc_data:
            idx = self.lc_data.sn == sn
            lc_data = self.lc_data.nt[idx]

        spec_data = None
        if self.spec_data:
            idx = self.spec_data.sn == sn
            spec_data = self.spec_data.nt[idx]

        spectrophotometric_data = None
        if self.spectrophotometric_data is not None:
            idx = self.spectrophotometric_data.sn == sn
            spectrophotometric_data = self.spectrophotometric_data.nt[idx]

        return TrainingDataset(sn_data,
                               lc_data=lc_data,
                               spec_data=spec_data,
                               spectrophotometric_data=spectrophotometric_data)

    # def init_model_plotter(self):
    #     return ModelPlotter(SALT2Like, self.get_all_filter_names())

    def copy(self):
        sn_data = self.sn_data.nt.copy()

        lc_data, spec_data, spectrophotometric_data = None, None, None
        if self.lc_data:
            lc_data = self.lc_data.nt.copy()
        if self.spec_data:
            spec_data = self.spec_data.nt.copy()
        if self.spectrophotometric_data:
            spectrophotometric_data = self.spectrophotometric_data.nt.copy()

        return TrainingDataset(sn_data, lc_data,
                               spec_data, spectrophotometric_data)

    def _select(self, selectors=None):
        """perform a data selection, before indexation.

        We may want to perform some a-priori selection, before the dataset is
        built and indexed. Re-indexation after selection is a little costly
        (seconds for O(10^4) SNe, dozens of seconds for a LSST-like sample).

        For this reason, this selection is done only once, when building the
        dataset. Outlier removal is then done by maintaining a `valid` field
        in the spec_data and lc_data arrays.
        """
        for select in selectors:
            select(self)

    def kill_sne(self, sn_list):
        """kill the SNe listed in argument

        invalidate the SNe listed in argument, along with their associated
        follow-up data points (light curves and spectra). This method only
        modified the `valid` field in self.sne, self.lc_data and
        self.spec_data.  Call `self.compress` to effectively remove the data
        and re-index the dataset.

        Parameters
        ----------
        sn_list : List[int]
            the ids of the supernovae to remove
        """
        for sn in sn_list:
            # invalidate the SN
            idx = self.sn_data.sn == sn
            if idx.sum() == 0:
                logging.warning(f'kill_sne: no SN with index {sn}')
                continue
            if idx.sum() > 1:
                raise TrainingDatasetError(f'more than one SN with index {sn}')
            self.sn_data.valid[idx] = 0
            # the associated light curves
            if self.lc_data:
                self.lc_data.valid[self.lc_data.sn == sn] = 0
            # and the associated spectra
            if self.spec_data:
                self.spec_data.valid[self.spec_data.sn == sn] = 0
            if self.spectrophotometric_data:
                self.spectrophotometric_data.valid[self.spectrophotometric_data.sn == sn] = 0

    def kill_spectra(self, spec_list):
        """kill the spectra listed in argument
        """
        if self.spec_data is None:
            logging.warning('kill_spectra: no spectra in this dataset')
            return
        for spec in spec_list:
            idx = self.spec_data.spec == spec
            if idx.sum() == 0:
                logging.warning(f'no spectral data for index {spec}')
                continue
            self.spec_data.valid[idx] = 0

    def kill_photometric_spectra(self, spec_list):
        """kill the spectra listed in argument
        """
        if self.spectrophotometric_data is None:
            logging.warning('kill_photometric_spectra: no spectra in this dataset')
            return
        for spec in spec_list:
            idx = self.spectrophotometric_data.spec == spec
            if idx.sum() == 0:
                logging.warning(f'no spectral data for index {spec}')
                continue
            self.spectrophotometric_data.valid[idx] = 0

    def kill_lcs(self, lc_list):
        """kill the light curves with the specified indexes

        Parameters
        ----------
        lc_list : List[int]
            the indexes of the light curves to kill
        """
        if self.lc_data is None:
            logging.warning('kill_lcs: no spectra in this dataset')
            return
        for lc in lc_list:
            idx = self.lc_data.lc == lc
            if idx.sum() == 0:
                logging.warning(f'no light curve points for index {lc}')
                continue
            self.lc_data.valid[idx] = 0

    def compress(self):
        """get rid of all the data that has been invalidated

        If a large amount of data has been flagged as invalid, then it may make
        sense to get rid of it. This operation is costly however, as the data
        is copied and the entire index needs to be rebuilt.  """
        nlc, nsp, nspp = self.nb_meas(valid_only=True, split_by_type=True)

        self.sn_data.compress(self.sn_data.valid==1)
        if nlc > 0:
            self.lc_data.compress(self.lc_data.valid==1)
        else:
            self.lc_data = None

        if nsp > 0:
            self.spec_data.compress(self.spec_data.valid==1)
        else:
            self.spec_data = None

        if nspp > 0:
            self.spectrophotometric_data.compress(self.spectrophotometric_data.valid==1)
        else:
            self.spectrophotometric_data = None

        self._index_sne()
        if self.lc_data:
            self._index_lcs()
        if self.spec_data:
            self._index_spectra()
        if self.spectrophotometric_data:
            self._index_photometric_spectra()

        # build a global index. it is costly, but it can help
        # self.sne, self.lcs, self.spectra = build_index(self)

    def _index_sne(self):
        """
        """
        # contiguous index for SNe
        self.sn_data.make_index('sn')
        # row number of each SN
        self.isn = dict([(self.sn_data.sn_index[i], i) for i in range(len(self.sn_data))])

    def _index_lcs(self):
        """Sort and index the light curve data

        Sort the light curve data, so that the light curve measurements are
        stored in continuous chunks and can be indexed easily.

        Build an index for each light curve. Here, a light curve is the set of
        all photometric points, for one sn, measured in one band). The index
        entries contains a slice object and links to the original data.
        """
        logging.info('indexing light curves')
        if not self.lc_data:
            logging.warning('_index_lcs: no lc data')
            return

        self.lc_data.make_index('lc')
        self.lc_data.make_index('band')
        # we need to recycle the SN index generated for sn_data
        # otherwise, we may get inconsistant sn_indexes between
        # the three structures
        sn_index = np.array([self.sn_data.sn_map[sn] for sn in self.lc_data.sn])
        self.lc_data.add_field('sn_index', sn_index)

        # add a pointer to the SN row
        if not hasattr(self, 'isn'):
            self._index_sne()
        nsn = len(self.isn)
        lc_isn = np.array([self.isn[idx] for idx in self.lc_data.sn_index])
        self.lc_data.add_field('isn', lc_isn)

        # since we keep using the SN redshift in the model evaluations,
        # add it to the LC data
        self.lc_data.add_field('z', self.sn_data.z[self.lc_data.isn])

        # we also need a field to identify the row number. It is used a lot
        # when building the model jacobian matrices. With DataFrames, we would
        # have an index. Here, we need to build one ourselves.
        self.lc_data.add_field('row', np.arange(len(self.lc_data)))

        # array pointing to the first row of each light curve
        # useful to handle per-lightcurve information
        _, self.i_lc_first = np.unique(self.lc_data.lc_index, return_index=True)

        # create a small lc index
        self.lc_db = np.unique(np.rec.fromarrays((self.lc_data.lc_index,
                                                  self.lc_data.sn_index,
                                                  self.lc_data.band,
                                                  self.lc_data.z), \
                                                 names=['lc_index', 'sn_index', 'band', 'z']))


    def _index_spectra(self):
        """Sort and index the spectral data

        Sort the spectral data, so that the spectrum measurements are stored in
        contiguous chunks and can be indexed easily.

        Build an index for each spectrum. An index entry contains a slice object
        and links to the original data.

        """
        logging.info('indexing spectra')
        if not hasattr(self, 'spec_data') or not self.spec_data:
            logging.warning('_index_spectra: no spectra no index - pass')
            return

        self.spec_data.make_index('spec')
        # we need to recycle the SN index generated by the sn_data proxy
        sn_index = np.array([self.sn_data.sn_map[sn] for sn in self.spec_data.sn])
        self.spec_data.add_field('sn_index', sn_index)

        # add a pointer to each SN row
        if not hasattr(self, 'isn'):
            self._index_sne()
        spec_isn = np.array([self.isn[idx] for idx in self.spec_data.sn_index])
        self.spec_data.add_field('isn', spec_isn)

        # add the redshift, since we keep using it when evaluating the model
        self.spec_data.add_field('z', self.sn_data.z[self.spec_data.isn])

        # we need a field to identify the row number. It is used a lot when
        # building the model jacobian matrix. With pandas.DataFrames, we would
        # have an index. Here, we need to build it ourselves.
        nlc, nspec, nphotspec = self.nb_meas(valid_only=False,
                                             split_by_type=True)
        self.spec_data.add_field('row', nlc + np.arange(len(self.spec_data)))

        # create a small spectrum index
        self.spec_db = np.unique(np.rec.fromarrays((self.spec_data.spec_index,
                                                    self.spec_data.mjd,
                                                    self.spec_data.sn_index,
                                                    self.spec_data.z), \
                                                   names=['spec_index', 'mjd', 'sn_index', 'z']))

    def _index_photometric_spectra(self):
        """Same, for the photometric spectra (a.k.a. photometric dataset)

        Sort the spectral data, so that the spectrum measurements are stored
        contiguously and can be indexed easily.

        For now, we do not build an index (as list of SpectrumData objects) for
        each spectrum. This is not necessary, as the model evaluation is
        totally vectorized (this is not the :wcase for the light curves, which
        is why we still build a separate index).
        """
        logging.info('indexing photometric (calibrated) spectra')
        if not hasattr(self, 'spectrophotometric_data') or not self.spectrophotometric_data:
            logging.warning('_index_photometric_spectra: no data to index - pass')
            return

        # index the spectra
        self.spectrophotometric_data.make_index('spec')
        sn_index = np.array([self.sn_data.sn_map[sn] for sn in self.spectrophotometric_data.sn])
        self.spectrophotometric_data.add_field('sn_index', sn_index)

        # add a pointer to each SN row
        if not hasattr(self, 'isn'):
            self._index_sne()
        spec_isn = np.array([self.isn[idx] for idx in self.spectrophotometric_data.sn_index])
        self.spectrophotometric_data.add_field('isn', spec_isn)

        # add the redshift, since we keep using it when evaluating the model
        self.spectrophotometric_data.add_field('z', self.sn_data.z[self.spectrophotometric_data.isn])

        # we need a field to identify the row number. It is used a lot when
        # building the model jacobian matrix. With pandas.DataFrames, we would
        # have an index. Here, we need to build it ourselves.
        nlc, nspec, nphotspec = self.nb_meas(valid_only=False, split_by_type=True)
        self.spectrophotometric_data.add_field('row', nlc + nspec + np.arange(nphotspec))

        # create a small spectrum index
        self.spectrophot_db = np.unique(np.rec.fromarrays((self.spectrophotometric_data.spec_index,
                                                           self.spectrophotometric_data.mjd,
                                                           self.spectrophotometric_data.sn_index,
                                                           self.spectrophotometric_data.z), \
                                                          names=['spec_index', 'mjd', 'sn_index', 'z']))

    def __len__(self):
        """number of data points (spec+phot, valid+invalid)"""
        return self.nb_meas(valid_only=False, split_by_type=False)

    def nb_meas(self, valid_only=True, split_by_type=False):
        """number of measurements, (spec+phot)"""
        if valid_only:
            n_phot_meas = int(self.lc_data.valid.sum()) if self.lc_data is not None else 0
            n_spec_meas = int(self.spec_data.valid.sum()) if self.spec_data is not None else 0
            n_spectrophot_meas = int(self.spectrophotometric_data.valid.sum()) if self.spectrophotometric_data is not None else 0
        else:
            n_phot_meas = len(self.lc_data) if self.lc_data is not None else 0
            n_spec_meas = len(self.spec_data) if self.spec_data is not None else 0
            n_spectrophot_meas = len(self.spectrophotometric_data) if self.spectrophotometric_data is not None else 0

        if split_by_type:
            return n_phot_meas, n_spec_meas, n_spectrophot_meas
        return n_phot_meas + n_spec_meas + n_spectrophot_meas

    def concat(self):
        """return a concatenation of all data
        """
        pass

    def nb_sne(self, valid_only=True):
        """
        Return number of (valid) SNe.
        """
        if valid_only:
            return int(self.sn_data.valid.sum())
        return len(self.sn_data)

    def nb_lcs(self, valid_only=True):
        """
        Return number of (valid) light curves.

        .. todo:: implement valid_only
        """
        if not hasattr(self, 'lc_data') or not self.lc_data:
            return 0

        if valid_only:
            idx = self.lc_data.valid > 0
            return len(np.unique(self.lc_data.lc[idx]))
        return len(np.unique(self.lc_data.lc))

    def nb_spectra(self, valid_only=True):
        """
        Return number of spectra.

        .. todo:: implement valid_only
        """
        if not hasattr(self, 'spec_data') or not self.spec_data:
            return 0

        if valid_only:
            idx = self.spec_data.valid > 0
            return len(np.unique(self.spec_data.spec[idx]))
        return len(np.unique(self.spec_data.spec))

    def nb_photometric_spectra(self, valid_only=True):
        """Return number of photometric spectra
        (i.e. SNfactory like spectra)
        """
        if not hasattr(self, 'spectrophotometric_data') or not self.spectrophotometric_data:
            return 0

        if valid_only:
            idx = self.spectrophotometric_data.valid > 0
            return len(np.unique(self.spectrophotometric_data.spec[idx]))
        return len(np.unique(self.spectrophotometric_data.spec))

    def nb_bands(self):
        return len(self.transmissions)

    def update_fluxes(self, flx):
        """update the fluxes of all data blocks
        """
        nlc, nsp, nspphot = self.nb_meas(valid_only=False,
                                         split_by_type=True)
        if nlc > 0:
            self.lc_data.flux[:] = flx[:nlc]
        if nsp > 0:
            self.spec_data.flux[:] = flx[nlc:nlc+nsp]
        if nspphot > 0:
            self.spectrophotometric_data.flux[:] = flx[nlc+nsp:]

    def get_all_fluxes(self):
        blocks = []
        for block in [self.lc_data, self.spec_data,
                      self.spectrophotometric_data]:
            if block is not None:
                blocks.append(block.flux)
        assert(len(blocks) > 0)
        return np.hstack(blocks)

    # def set_fluxes(self, all_fluxes=None, **kwargs):
    #     if all_fluxes is not None:
    #         nlc, nspec, _ = self.nb_meas(valid_only=False,
    #                                      split_by_type=True)
    #         if self.lc_data:
    #             self.lc_data.flux[:] = all_fluxes[0:nlc]
    #         if self.spec_data:
    #             self.spec_data.flux[:] = all_fluxes[nlc:nlc+nspec]
    #         if self.spectrophotometric_data:
    #             self.spectrophotometric_data[:] = all_fluxes[nlc+nspec:]

    def get_all_fluxerr(self):
        blocks = []
        for block in [self.lc_data, self.spec_data,
                      self.spectrophotometric_data]:
            if block is not None:
                blocks.append(block.fluxerr)
        assert(len(blocks) > 0)
        return np.hstack(blocks)

    def get_valid(self):
        blocks = []
        for block in [self.lc_data, self.spec_data,
                      self.spectrophotometric_data]:
            if block is not None:
                blocks.append(block.valid)
        assert(len(blocks) > 0)
        return np.hstack(blocks)

    def reset_valid(self, valid=1):
        for block in [self.lc_data, self.spec_data,
                      self.spectrophotometric_data]:
            if block is not None:
                block.valid[:] = valid

    def get_bads(self):
        blocks = []
        for block in [self.lc_data, self.spec_data,
                      self.spectrophotometric_data]:
            if block is not None:
                blocks.append(block.valid==0)
        assert(len(blocks) > 0)
        return np.hstack(blocks)

    def get_sn_pars(self, sn):
        """retrieve the sn pars from the training dataset (if available)
        """
        ret = {}
        idx = self.sn_data.sn == sn
        sn_info = self.sn_data.nt[idx]
        keys = sn_info.dtype.names
        for k in ['z', 'x0', 'x1', 'c', 'tmax', 'ebv']:
            ret[k] = float(sn_info[k]) if k in keys else 0.
        return ret

    def get_all_filter_names(self, force=False) -> List[str]:
        """passbands used in the dataset

        scan the dataset and return a list containing the full names
        (`instrument`::`band_name`) of all the passbands that were used in the
        photometric follow-up.

        Returns
        -------
        List[str]
            List of filter names
        """
        if not hasattr(self, 'lc_data') or not self.lc_data:
            return None

        if hasattr(self, 'filter_names') and not force:
            return self.filter_names
        t = np.unique(self.lc_data.band)
        self.filter_names = t.astype(t.dtype.str.replace('S', 'U'))
        return self.filter_names
    
    
    def get_all_transmissions(self, force=False):
        """transmissions used in the dataset

        from the list of passbands (see above) fetch all the corresponding
        transmissions and return them.

        Returns
        -------
        list
            List of filter transmission.

        """
        if not hasattr(self, 'filterlib') or self.filterlib is None:
            logging.warning('unable to load transmissions: no filterlib')
            return

        if not hasattr(self, 'lc_data') or not self.lc_data:
            logging.warning('no transmissions to load: no photometric data')
            return

        if hasattr(self, 'transmissions') and not force:
            return self.transmissions
        self.transmissions = dict([(fn, self.filterlib[fn]) for fn in self.get_all_filter_names(force=force)])

        return self.transmissions

    def get_zp(self, sn):
        """
        """
        idx = self.lc_data.sn == sn
        return dict(zip(self.lc_data.band[idx], self.lc_data.zp[idx]))

    def get_magsys(self, sn):
        """
        """
        idx = self.lc_data.sn == sn
        return dict(zip(self.lc_data.band[idx], self.lc_data.magsys[idx]))

    # def get_magsys_zp(self, force=False):
    #     """
    #     """
    #     if not hasattr(self, 'lc_data') or not self.lc_data:
    #         return

    #     if hasattr(self, 'magsys_zp') and not force:
    #         return self.magsys_zp

    #     try:
    #         # if force, we need to update the transmissions
    #         self.get_all_transmissions(force=force)

    #         lc_data = self.lc_data
    #         self.magsys = dict([(magsys_name, MagSys(magsys_name)) for magsys_name in np.unique(lc_data.magsys)])

    #         self.magsys_zp = {}
    #         for magsys_name, band_name in set(zip(lc_data.magsys, lc_data.band)):
    #             ms = self.magsys[magsys_name]
    #             tr = self.transmissions[band_name]
    #             self.magsys_zp[(magsys_name, band_name)] = ms.ZeroPoint(tr)
    #         zp = np.array([self.magsys_zp[(m,b)] for m,b in zip(lc_data.magsys, lc_data.band)])
    #         lc_data.add_field('magsys_zp', zp)


    #     except:
    #     # if force, we need to update the transmissions
    #         self.get_all_transmissions(force=force)

    #         m_AB = sncosmo.ABMagSystem()
    #         self.magsys_zp ={}

    #         lc_data = self.lc_data
    #         self.magsys = dict([(magsys_name, m_AB) for magsys_name in np.unique(lc_data.magsys)])

    #         self.magsys_zp = {}
    #         for magsys_name, band_name in set(zip(lc_data.magsys, lc_data.band)):
    #             ms = self.magsys[magsys_name]
    #             tr = self.transmissions[band_name]
    #             band = sncosmo.get_bandpass(band_name)
    #             self.magsys_zp[(magsys_name, band_name)] =  -2.5*np.log10(m_AB.zpbandflux(band)) #ms.ZeroPoint(tr)
    #         zp = np.array([self.magsys_zp[(m,b)] for m,b in zip(lc_data.magsys, lc_data.band)])
    #         lc_data.add_field('magsys_zp', zp)

    # def compute_photometric_norm_factors(self, force=False):
    #     """
    #     """
    #     if not hasattr(self, 'lc_data') or not self.lc_data:
    #         return

    #     # self.get_magsys_zp(force=force)

    #     lc_data = self.lc_data
    #     meas_zp_norm = 10**(-0.4 * lc_data.zp)
    #     # magsys_zp_norm = 10**(-0.4 * lc_data.magsys_zp)  # was -0.4
    #     norm = meas_zp_norm # * magsys_zp_norm
    #     lc_data.add_field('norm', norm)

    def update_zp_scale(self):
        """derive a multiplicative calibration scale from the zero point
        """
        if not hasattr(self, 'lc_data') or not self.lc_data:
            return

        scale = 10.**(-0.4 * self.lc_data.zp)
        self.lc_data.add_field('zp_scale', scale)

    def to_parquet(self, name, path=pathlib.Path('./')):
        """Write the internal dataframes to the binary parquet format

        This method writes each of the internal array (sn index, lightcurves,
        spectra, photometric spectra) as parquet files (one for each).

        Parameters
        ----------
        name : str or path object
          core file name. Each file is written as <name>_<suffix>.parquet
        path : str or path object
          root path
        """
        path = pathlib.Path(path)
        fn = path.joinpath(name)
        
        # save the supernova metadata
        df = pandas.DataFrame(self.sn_data.nt).set_index('index')
        df.to_parquet(fn.with_suffix(".sn.parquet"),
                      engine='pyarrow', index=False)
        # save the lc data, if any
        if hasattr(self, 'lc_data') and self.lc_data is not None:
            df = pandas.DataFrame(self.lc_data.nt).set_index('index')
            df.to_parquet(fn.with_suffix(".lc.parquet"),
                          engine='pyarrow', index=False)
        # save the spectra, if any
        if hasattr(self, 'spec_data') and self.spec_data is not None:
            df = pandas.DataFrame(self.spec_data.nt).set_index('index')
            df.to_parquet(fn.with_suffix(".spec.parquet"),
                          engine='pyarrow', index=False)

        # save the spectrophotometric data, if available
        if hasattr(self, 'spectrophotometric_data') and \
           self.spectrophotometric_data is not None:
            df = pandas.DataFrame(self.spectrophotometric_data).set_index('index')
            df.to_parquet(fn.with_suffix(".specphot.parquet"),
                          engine='pyarrow', index=False)

        # save the basis - as a pkl file
        #        df = pandas.DataFrame({'basis': self.basis.grid})
        #        df.to_parquet(fn.with_suffix(".basis.parquet"), engine='pyarrow',
        #              index=False)
        if hasattr(self, 'basis') and self.basis is not None:
            with open(fn.with_suffix(".wlbasis.pkl"), 'wb') as f:
                pickle.dump(self.basis, f)

    @classmethod
    def read_parquet(cls, name, path='./', filterlib=None):
        """Load a TrainingDataset from parquet files

        Parameters
        ----------
        name : str or path-like object
          core file name. The function searches for parquet files
          named `<name>_<suffix>.parquet` with suffix equal to `_sn`,
          `_lc`, `_spec`, `_specphot` for each component.
        path : str or path object
          root path
        """
        path = pathlib.Path(path)
        fn = path.joinpath(name)
        sn_data = pandas.read_parquet(fn.with_suffix(".sn.parquet"))

        # light curves ?
        lc_path = fn.with_suffix(".lc.parquet")
        if lc_path.is_file():
            lc_data = pandas.read_parquet(lc_path)
        else:
            lc_data = None

        # spectra ?
        spec_path = fn.with_suffix(".spec.parquet")
        if spec_path.is_file():
            spec_data = pandas.read_parquet(spec_path)
        else:
            spec_data = None

        # spectrophotometric data ?
        # spectra ?
        specphot_path = fn.with_suffix(".specphot.parquet")
        if specphot_path.is_file():
            specphot_data = pandas.read_parquet(specphot_path)
        else:
            specphot_data = None

        # wavelength basis (if available)
        basis_path = fn.with_suffix(".wlbasis.pkl")
        if basis_path.is_file():
            with open(basis_path, 'rb') as f:
                basis = pickle.load(f)
        else:
            basis = None                

        print(basis_path, basis_path.is_file(), basis)
            
        return cls(sn_data, lc_data=lc_data, spec_data=spec_data,
                   spectrophotometric_data=specphot_data,
                   filterlib=filterlib, basis=basis)

    def to_npz(self, path, compressed=False):
        """Write the TrainingDataset instance as a numpy npz file
        """
        path = pathlib.Path(path)
        if compressed:
            np.savez_compressed(path, sn_data=self.sn_data,
                                lc_data=self.lc_data,
                                spec_data=self.spec_data)
        else:
            np.savez(path, sn_data=self.sn_data,
                     lc_data=self.lc_data,
                     spec_data=self.spec_data)

    @classmethod
    def read_npz(cls, path):
        """Load a TrainingDataset from a numpy npz file
        """
        path = pathlib.Path(path)
        f = np.load(path, allow_pickle=True)
        lc_data = f['lc_data'] if 'lc_data' in f else None
        spec_data = f['spec_data'] if 'spec_data' in f else None
        specphot_data = f['specphot_data'] if 'specphot_data' in f else None
        return cls(f['sn_data'], lc_data=lc_data, spec_data=spec_data,
                   spectrophotometric_data=specphot_data)

    def to_hdf(self, path, compressed=True):
        """Write the training dataset in a HDF5 file (binary format)
        """
        path = pathlib.Path(path)
        with pandas.HDFStore(path, complevel=9) as f:
            f.put('sn_data',
                  pandas.DataFrame(self.sn_data.nt).set_index('index'))
            if hasattr(self, 'lc_data') and self.lc_data:
                f.put('lc_data',
                      pandas.DataFrame(self.lc_data.nt).set_index('index'))
            if hasattr(self, 'spec_data') and self.spec_data:
                f.put('spec_data',
                      pandas.DataFrame(self.spec_data.nt).set_index('index'))
            if hasattr(self, 'spectrophotometric_data') and self.spectrophotometric_data:
                f.put('spectrophotometric_data',
                      pandas.DataFrame(self.spectrophotometric_data.nt).set_index('index'))
            f.put('basis',
                  pandas.DataFrame({'basis':self.basis}, index=[0]))

    @classmethod
    def read_hdf(cls, path):
        """Load a training dataset from a HDF5 file
        """
        path = pathlib.Path(path)
        with pandas.HDFStore(path) as f:
            lc_data = f.lc_data if 'lc_data' in f else None
            spec_data = f.spec_data if 'spec_data' in f else None
            spectrophotometric_data = f.spectrophotometric_data if 'spectrophotometric_data' in f else None
            basis_ = f.basis['basis'][0] if 'basis' in f else None
            return cls(f.sn_data, lc_data=lc_data, spec_data=spec_data,
                      spectrophotometric_data=spectrophotometric_data,
                      basis = basis_)

    def plot_sample(self, bins=50):
        """plot a synthetic view of the sample

        this method displays, in a single panel:
            - a histogram of the redshifts
            - histograms of X1 and color
            - the variations of X0 as a function of redshift
        """
        fig, axes = plt.subplots(2,2, figsize=(8,8))
        fig.suptitle(f'sample: size={len(self.sn_data)}')
        axes[0, 0].hist(self.sn_data.z, bins=bins, alpha=0.5)
        axes[0, 0].set_xlabel('$z$')

        axes[1, 0].semilogy(self.sn_data.z, self.sn_data.x0,
                           color='k', ls='', marker='.')
        axes[1, 0].set_xlabel('$z$')
        axes[1, 0].set_ylabel('$X_0$')

        axes[0, 1].hist(self.sn_data.x1, bins=bins, alpha=0.5)
        axes[0, 1].set_xlabel('$X_1$')
        axes[1, 1].hist(self.sn_data.col, bins=bins, alpha=0.5)
        axes[1, 1].set_xlabel('$c$')

    def plot_lcs(self, sn, model=None, pars=None,
                 color=None, cmap=None, cmap_lims=(3000., 10000.)):
        """plot the light curves of the specified SN

        Parameters:
        -----------
        sn : (int)
            unique sn id in the training dataset
        model : (SALT2Eval)
            a model eval class (the thing that compute the model a-la sncosmo)
        pars : (FitParameters)
            additional parameters for the model
        """
        idx = self.sn_data.sn == sn
        tmax = float(self.sn_data.tmax[idx])

        idx = self.lc_data.sn == sn
        lc_bands = sorted(np.unique(self.lc_data.band[idx]).tolist(),
                          key=opt_band_name_order)
        nb_bands = len(lc_bands)
        ncols = int(np.sqrt(nb_bands))
        nrows = int(nb_bands/ncols + 1)

        # color & color map
        # I am copying sncosmo's logic here
        if color is None:
            if cmap is None:
                cmap = cm.get_cmap('jet_r')

        # overploting a model
        if model is not None:
            model.set_from_tds(sn, self)
            if pars is not None:
                model.update_global_pars(pars)
            model()

        fig = plt.figure(figsize=(14, 10))
        z = np.unique(self.lc_data.z[idx])[0]
        fig.suptitle(f'SN #{sn}  [z={z}]')

        for i, b in enumerate(lc_bands):
            # axis
            p, q = int(i / ncols), i % ncols
            ax = fig.add_subplot(nrows, ncols, i+1)

            # color
            if color is None:
                wl = self.transmissions[b].wave_eff()
                bandcolor = cmap((wl-cmap_lims[0]) /
                             (cmap_lims[1] - cmap_lims[0]))
            else:
                bandcolor = color

            # plot all the LC points
            idx = (self.lc_data.sn == sn) & (self.lc_data.band == b)
            plt.plot(self.lc_data.mjd[idx], self.lc_data.flux[idx],
                    ls='', marker=',', color=bandcolor)

            # then, just the valid points
            idx = (self.lc_data.sn == sn) & (self.lc_data.band == b) & (self.lc_data.valid == 1)
            plt.errorbar(self.lc_data.mjd[idx], self.lc_data.flux[idx],
                        yerr=self.lc_data.fluxerr[idx],
                        ls='', marker='.', color=bandcolor)

            # and all the points that were killed and are no longer
            # part of the analysis
            idx_killed = (self.lc_data.sn == sn) & (self.lc_data.band == b) & (self.lc_data.valid == 0)
            plt.plot(self.lc_data.mjd[idx_killed], self.lc_data.flux[idx_killed],
                    ls='', marker='x', color=bandcolor)

            # if model requested, we plot it on top of the data
            if model:
                idx = model.tds.lc_data.band == b
                lc_data = model.tds.lc_data.nt[idx]
                ax.plot(lc_data.mjd, lc_data.flux, ls='-', color=bandcolor)
            ax.set_ylabel(f'$flux_{b}$')
            ax.axvline(tmax)

    def plot_spectra(self, sn, model=None):
        """plot all the spectra of the specified SN
        """
        idx = self.sn_data.sn == sn
        tmax = float(self.sn_data.tmax[idx])
        pass

    def plot_spectrum(self, spec, phot=False, model=None, pars=None,
                      color='b'):
        """plot the spectrum number spec
        """
        if not phot:
            spec_data = self.spec_data
        else:
            spec_data = self.spectrophotometric_data
        assert spec_data is not None

        idx = spec_data.spec == spec
        sn = spec_data.sn[idx][0]
        mjd = spec_data.mjd[idx][0]
        z = spec_data.z[idx][0]
        logging.info(f'{sn} {mjd} {z}')

        # overplotting a model
        if model is not None:
            model.set(spec_mjd=mjd)
            model.set_from_tds(sn, self)
            if pars is not None:
                model.update_global_pars(pars)
            model()

        fig = plt.figure(figsize=(14,10))
        ax = fig.add_subplot(111)
        fig.suptitle(f'spectrum {spec}' +
                     f' [SN: {sn}],' +
                     f' z={z:.3f}')
        ax.plot(spec_data.wavelength[idx], spec_data.flux[idx], 'k,')
        idx = (spec_data.spec == spec) & (spec_data.valid == 1)
        ax.errorbar(spec_data.wavelength[idx], spec_data.flux[idx],
                    yerr=spec_data.fluxerr[idx], ls='', marker='.', color='k')
        idx_killed = (spec_data.spec == spec) & (spec_data.valid==0)
        ax.plot(spec_data.wavelength[idx_killed], spec_data.flux[idx_killed],
                color='r', marker='x', ls='')

        if model:
            if not phot:
                spec_data = model.tds.spec_data.nt
            else:
                spec_data = model.tds.spectrophotometric_data.nt

            ax.plot(spec_data.wavelength, spec_data.flux,
                    ls='-', color=color)
            print(spec_data.flux)
        # if model:
        #     if self.plotter is None:
        #         self.plotter = self.init_model_plotter()
        #     sn = spec_data.sn[idx][0]
        #     mjd = spec_data.mjd[idx][0]
        #     assert np.all(spec_data.sn[idx] == sn)
        #     assert np.all(spec_data.mjd[idx] == mjd)
        #     self.plotter.load_sn_pars(sn, self, spec_mjd=mjd, pars=pars)
        #     self.plotter.plot_spectrum(ax=ax, phot=phot)

        ax.set_xlabel('$\lambda\ \ [\AA]$')
        ax.set_ylabel('flux')

    def plot_coverage(self, log=False):
        """$\lambda$-phase plane coverage
        """
        wl_bins = np.linspace(2000, 10000, 120)
        phase_bins = np.linspace(-20, 50, 40)

        # we plot things differently according to whether
        # there are two or just one spectroscopic dataset
        if self.spec_data is None and self.spectrophotometric_data is None:
            fig, axes = plt.subplots(nrows=2, ncols=2,
                                    sharex=True, sharey=True,
                                    figsize=(14,14))
            lc_axes = axes[0, 0] if self.lc_data else None
            spec_axes = axes[0, 1] if self.spec_data else None
            photspec_axes = axes[1,0] if self.spectrophotometric_data else None
            sum_axes = axes[1,1] if self.spec_data and \
                self.spectrophotometric_data else None
        else:
            fig, axes = plt.subplots(nrows=1, ncols=2,
                                    sharex=True, sharey=True,
                                    figsize=(14,7))
            lc_axes = axes[0] if self.lc_data else None
            spec_axes = axes[1] if self.spec_data else None
            photspec_axes = axes[1] if self.spectrophotometric_data else None
            sum_axes = None

        H_lc, H_spec, H_spectrophot = None, None, None

        # LC coverage
        if lc_axes is not None:
            zz = 1. + self.lc_data.z
            wl = self.lc_data.wavelength / zz
            tmax = self.sn_data.tmax[self.lc_data.isn]
            phase = (self.lc_data.mjd - tmax) / zz
            norm = matplotlib.colors.LogNorm() if log else None
            H_lc, _, _, im = lc_axes.hist2d(wl, phase,
                                            bins=(wl_bins, phase_bins),
                                            norm=norm)

            lc_axes.set_title('photometric coverage')
            lc_axes.set_xlabel('$\lambda$ [restframe, $\AA$]')
            lc_axes.set_ylabel('phase [restframe days]')
            plt.colorbar(im)

        # spectral coverage
        if spec_axes is not None:
            zz = 1. + self.spec_data.z
            wl = self.spec_data.wavelength / zz
            tmax = self.sn_data.tmax[self.spec_data.isn]
            phase = (self.spec_data.mjd - tmax) / zz
            H_spec, _, _, im = spec_axes.hist2d(wl, phase,
                                               bins=(wl_bins, phase_bins))
            spec_axes.set_title('spectral coverage')
            spec_axes.set_xlabel('$\\lambda$ [restframe, $\AA$]')
            spec_axes.set_ylabel('phase [restframe days]')
            plt.colorbar(im)

        if photspec_axes is not None:
            zz = 1. + self.spectrophotometric_data.z
            wl = self.spectrophotometric_data.wavelength / zz
            tmax = self.sn_data.tmax[self.spectrophotometric_data.isn]
            phase = (self.spectrophotometric_data.mjd - tmax) / zz
            H_spectrophot, _, _, im = photspec_axes.hist2d(wl, phase,
                                                 bins=(wl_bins, phase_bins))
            photspec_axes.set_title('photometric spectra')
            photspec_axes.set_xlabel('$\\lambda$ [restframe, $\AA$]')
            photspec_axes.set_ylabel('phase [restframe days]')
            plt.colorbar(im)

        if sum_axes is not None:
            I = sum_axes.matshow(H_spec + H_spectrophot, aspect='auto')
            sum_axes.set_title('total spectral coverage')
            sum_axes.set_xlabel('$\\lambda$ [restframe, $\AA$]')
            sum_axes.set_ylabel('phase [restframe days]')
            plt.colorbar(I)

        plt.subplots_adjust(wspace=0.1, left=0.1, right=0.95)


def _df_to_recarray(x, sort_by=None):
    """utility function to convert the data into a sorted recarray

    The TrainingDataset expects either a recarray or a dataframe. Internally,
    it uses a recarrays, because accessing their content is about factor 10
    faster than pandas.DataFrames. It also requires the recarray to be sorted
    to speed up data access and allow easy indexing of the data.

    This function (optionally) sorts the input data according to the sort_by
    argument and returns the sorted data as a numpy.recarray.

    Parameters
    ----------
    x : pandas.DataFrame | numpy.recarray
        the data
    sort_by : List[str]], optional
        sort directives, by default None

    Returns
    -------
    numpy.recarray
        the (optionally sorted) data

    Raises
    ------
    ValueError
        if the input data is neither a pandas.DataFrame or a numpy.recarray
    """
    if isinstance(x, pandas.DataFrame):
        if sort_by:
            x = x.sort_values(by=sort_by)
        return x.to_records()

    if isinstance(x, np.recarray):
        if sort_by:
            x.sort(order=sort_by)
        return x

    raise ValueError('array is neither a DataFrame nor a recarray')

class SNData:

    def __init__(self, sn_data):
        """constructor

        Parameters
        ----------
        sn_data : record
            a record that contains at least the following fields:
              - 'z', the SN redshift (float)
              - 'sn', a unique ID for the SN (int or str)
              - 'valid', int
        """
        self.data = sn_data
        self.lcs = {}
        self.spectra = []

    @property
    def sn(self):
        return self.data.sn

    @property
    def z(self):
        return self.data.z

    @property
    def valid(self):
        return self.data.valid

    def kill(self):
        """invalidate the SN and all the associated follow-up data
        """
        self.data.valid = 0
        for lc in self.lcs.values():
            lc.data.valid[:] = 0
        for sp in self.spectra:
            sp.data.valid[:] = 0

    def plot(self):
        """standard plot to present the SN data"""
        pass


class LcData:
    """A class to access the light curve data."""

    def __init__(self, sn, band, slc, lc_data):
        """Constructor

        Parameters
        ----------
        sn : int
            Index of sn.
        band : str
            Name of the Filter.
        slc : slc
            Index of the Light Curve data in the full photometric data.
        lc_data : nacl.lib.dataproxy.DataProxy
            Photometric data.
        sne : numpy.recarray
            Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
        """
        self.sn_info = sn
        # note: this masks the `band` field in self.lc_data
        # self.band = band
        self.slc = slc
        self.lc_data = lc_data

    def __len__(self):
        """number of light curve data points.
        """
        return len(self.lc_data.row[self.slc])

    def __getattr__(self, name):
        try:
            return self.lc_data.__dict__[name][self.slc]
        except KeyError:
            raise AttributeError(f"{type(self).__name__} has no attribute {name}")

    # @property
    # def valid(self):
    #     return self.lc_data.valid[self.slc]

    # @property
    # def mjd(self):
    #     return self.lc_data.mjd[self.slc]

    # @property
    # def flux(self):
    #     return self.lc_data.flux[self.slc]

    # @property
    # def fluxerr(self):
    #     return self.lc_data.fluxerr[self.slc]

    # @property
    # def data(self):
    #     """
    #     Return numpy.rec.array data file.
    #     """
    #     return self.lc_data[self.slc]

    # @property
    # def z(self):
    #     """redshift"""
    #     return self.lc_data.z[self.slc]

    def kill(self):
        """invalidate all the light curve measurements"""
        self.valid[:] = 0

    def plot(self, ax=None):
        """standard light curve plot"""
        plt.figure()
        x, y, ey = self.mjd, self.flux, self.fluxerr
        plt.errorbar(x, y, yerr=ey, ls='', color='k', marker='o')
        plt.xlabel('phase [days]')
        plt.ylabel('Flux')
        sn = self.sn_info.sn
        band = np.unique(self.band)[0]
        z = self.sn_info.z
        plt.title('SN#{} {} [$z={:5.3}]$'.format(sn, band, z))


class SpectrumData:
    """A class to access and plot spectra
    """
    def __init__(self, sn, spec, slc, sp_data):
        """Constructor

        Parameters
        ----------
        sn : SNData
            info about the SN
        spectrum : str or int
            Spectrum unique identifier
        slc : slc
            slice o
        sp_data : numpy.recarray or pandas.DataFrame
            spectral data
        """
        self.sn_info = sn
        self.spec = spec
        self.slc = slc
        self.sp_data = sp_data

    def __len__(self):
        """number of data points"""
        return len(self.data)

    def __getattr__(self, name):
        try:
            return self.sp_data.__dict__[name][self.slc]
        except KeyError:
            raise AttributeError(f"{type(self).__name__} has no attribute {name}")

    # @property
    # def data(self):
    #     """the spectrum data itself"""
    #     return self.sp_data[self.slc]

    # @property
    # def z(self):
    #     """Sn redshift"""
    #     return self.sn.z

    def kill(self):
        self.valid[:] = 0

    def plot(self):
        """standard control plot
        """
        plt.figure()
        x, y, ey = self.wavelength, self.flux, self.fluxerr
        plt.errorbar(x, y, yerr=ey, ls='', color='k')
        plt.xlabel(r'$\lambda [\AA]$')
        plt.ylabel('flux')
        plt.title('SN#{} [$z={:5.3}$]'.format(self.sn_info.sn, self.sn_info.z))

def _index_sne(tds):
    sne = dict([(r.sn, SNData(r)) for r in tds.sn_data.nt])
    return sne

def _index_lcs(tds, sne):
    """parse the lc_data and build the light curve index

    Parse the lc_data, build a LcData handle for each identified LC,
    and attach it to its SN

    Parameters
    ----------
    lc_data : nacl.lib.DataProxy
        the DataProxy that handles all the lightcurves
    sne : Dict[SNData]
        the index - each entry corresponds to a SN
    """
    lcs = []
    lc_data = tds.lc_data
    if not lc_data:
        return

    # we need to assume at this point that the indexes are up to date
#    if not hasattr(lc_data, 'sn_index'):
#        lc_data.make_index('sn')
#    if not hasattr(lc_data, 'band_index'):
#        lc_data.make_index('band')
    index = lc_data.sn_index * 100 + lc_data.band_index

    # detect the block edges
    i_slices = np.where(index[1:] - index[:-1])[0] + 1
    i_slices = np.hstack(([0], i_slices.repeat(2), [len(lc_data)])).reshape((-1,2))

    # now comes the slow part -- we build the handles to each lightcurve
    # logging.info(f'building the LcData handles {i_slices.shape[0]} to build')
    lc_data_nt = lc_data.nt
    for i in range(i_slices.shape[0]):
        slc = slice(*i_slices[i])
        r = lc_data_nt[slc.start]
        lcdata = LcData(sne[r.sn], r.band, slc, lc_data)
        lcs.append(lcdata)
        sne[r.sn].lcs[r.band] = lcdata
    return lcs

def _index_spectra(tds, sne):
    """parse the spec_data, identifies all the spectra

    Parameters
    ----------
    spec_data : _type_
        _description_
    sne : _type_
        _description_
    """
    spec_data = tds.spec_data
    if spec_data is None:
        return

    spectra = []

    # we need to assume at this point
    # that the indexes are up to date
    #    if not hasattr(spec_data, 'spec_index'):
    #        spec_data.make_index('spec')
    #    if not hasattr(spec_data, 'sn_index'):
    #        spec_data.make_index('sn')

    # logging.info('indexing the spectra')
    mjd = spec_data.mjd
    i_slices = np.where(mjd[1:] - mjd[:-1])[0] + 1
    i_slices = np.hstack([[0],
                          i_slices.repeat(2),
                          [len(spec_data)]]).reshape((-1,2))

    # now, indexing
    spec_data_nt = spec_data.nt
    for i in range(i_slices.shape[0]):
        slc = slice(*i_slices[i])
        r = spec_data_nt[slc.start]
        spdata = SpectrumData(sne[r.sn], r.spec, slc, spec_data)
        spectra.append(spdata)
        sne[r.sn].spectra.append(spdata)
    return spectra

def build_index(tds):
    sne = _index_sne(tds)
    lcs = _index_lcs(tds, sne)
    spectra = _index_spectra(tds, sne)
    return sne, lcs, spectra

class SimTrainingDataset(TrainingDataset):
    """simple training dataset, with one single SN and regular sampling

    Useful for plotting the model over data, or model components.
    """

    def __init__(self, bands=['SWOPE::B'],
                 phase_range=(-20., 50.), restframe_dt=1.,
                 n_spectra=1, n_phot_spectra=1, zp=30., magsys='AB', filterlib=None):
        """generate a similar dataset, with the requested bands and spectra
        """
        # sn data
        sne = np.rec.array(np.zeros(1, dtype=TrainingDataset.sn_data_dtype))
        sne['valid'] = 1

        # light curves
        if bands is not None and len(bands) > 0:
            lc_data = []
            ph_min, ph_max = phase_range
            N = int((ph_max-ph_min) / restframe_dt)
            mjd = np.linspace(ph_min, ph_max, N)
            block_size = len(mjd)
            for band in bands:
                block = np.zeros(block_size,
                                 dtype=TrainingDataset.lc_data_dtype)
                block['mjd'] = mjd
                block['band'] = band
                block['valid'] = 1
                block['magsys'] = magsys
                if type(zp) is float:
                    block['zp'] = zp
                elif type(zp) is dict:
                    block['zp'] = zp.get(band, 30.)
                else:
                    block['zp'] = 30.
                block['x'] = 1000.
                block['y'] = 1000.
                block['sensor_id'] = 12
                lc_data.append(block)
            lc_data = np.rec.array(np.hstack(lc_data))
        else:
            lc_data = None

        # copy the mjd vector
        if lc_data is not None:
            self.mjd = lc_data.mjd.copy()
        else:
            self.mjd = None

        # spectra
        spec_data = None
        if n_spectra > 0:
            spec_data = []
            for sp in range(n_spectra):
                wl = np.linspace(3000., 11000., 8000)
                block_size = len(wl)
                block = np.zeros(block_size,
                                 dtype=TrainingDataset.spec_data_dtype)
                block['wavelength'] = wl
                block['valid'] = 1
                spec_data.append(block)
            spec_data = np.rec.array(np.hstack(spec_data))

        # photometric spectra
        spectrophotometric_data = None
        if n_phot_spectra > 0:
            spectrophotometric_data = []
            for sp in range(n_phot_spectra):
                wl = np.linspace(3000., 11000., 8000)
                block_size = len(wl)
                block = np.zeros(block_size,
                                 dtype=TrainingDataset.spec_data_dtype)
                block['wavelength'] = wl
                block['valid'] = 1
                spectrophotometric_data.append(block)
            spectrophotometric_data = np.rec.array(np.hstack(spectrophotometric_data))

        super().__init__(sne, lc_data, spec_data, spectrophotometric_data, filterlib=filterlib)



class ModelPlotter:

    def __init__(self, model_type, bands,
                 phase_range=(-20., 50.),
                 wl_range=(3000., 11000.),
                 n_spectra=1, n_phot_spectra=1, magsys='AB',
                 init_from_salt2_file='salt2.npz'):
        """
        """
        self.model_type = model_type
        self.bands = bands
        self.phase_range = phase_range
        self.wl_range = wl_range
        self.n_spectra = n_spectra
        self.init_from_salt2_file = init_from_salt2_file
        self.tds = self.init_training_dataset(bands,
                                              phase_range,
                                              wl_range,
                                              n_spectra,
                                              n_phot_spectra,
                                              magsys)
        self.model_type = model_type
        # self.model = model_type(self.tds,
        #                         init_from_salt2_file=init_from_salt2_file)

    def init_training_dataset(self, band_names,
                              phase_range, wl_range,
                              n_spectra, n_phot_spectra,
                              magsys):
        """
        """
        # sn_data (1 SN)
        sne = np.rec.array(np.zeros(1, dtype=TrainingDataset.sn_data_dtype))
        sne['valid'] = 1

        # light curves
        lc_data = []
        mjd = np.linspace(phase_range[0], phase_range[1], 100)
        block_size = len(mjd)
        if band_names is not None:
            for band_name in band_names:
                block = np.zeros(block_size,
                                 dtype=TrainingDataset.lc_data_dtype)
                block['mjd'] = mjd
                block['band'] = band_name
                block['magsys'] = magsys
                block['valid'] = 1
                lc_data.append(block)
            lc_data = np.rec.array(np.hstack(lc_data))
            self.lc_mjd = lc_data['mjd'].copy()
        else:
            lc_data = None
            self.lc_mjd = None

        # spectra
        spec_data = []
        for sp in range(n_spectra):
            wl = np.linspace(3000., 11000., 8000)
            block_size = len(wl)
            block = np.zeros(block_size,
                             dtype=TrainingDataset.spec_data_dtype)
            block['wavelength'] = wl
            block['valid'] = 1
            spec_data.append(block)
        spec_data = np.rec.array(np.hstack(spec_data))

        # photometric spectra
        spectrophotometric_data = []
        for sp in range(n_phot_spectra):
            wl = np.linspace(3000., 11000., 8000)
            block_size = len(wl)
            block = np.zeros(block_size,
                             dtype=TrainingDataset.spec_data_dtype)
            block['wavelength'] = wl
            block['valid'] = 1
            spectrophotometric_data.append(block)
        spectrophotometric_data = np.rec.array(np.hstack(spectrophotometric_data))

        return TrainingDataset(sne, lc_data, spec_data, spectrophotometric_data,
                               load_filters=True)

    def __call__(self, p):
        """
        """
        vals = self.model(p)

        # split the values into phot/spec/spectrophot blocks
        # depending on what we have in the dataset
        # TODO: there should be a tds function to handle that
        phot_blksize, spec_blksize, spectrophot_blksize = 0, 0, 0
        if self.tds.lc_data is not None:
            phot_blksize = len(self.tds.lc_data.nt)
            self.phot_vals = vals[0:phot_blksize]
        if self.tds.spec_data is not None:
            spec_blksize = len(self.tds.spec_data.nt)
            if spec_blksize > 0:
                self.spec_vals = vals[phot_blksize:phot_blksize+spec_blksize]
        if self.tds.spectrophotometric_data is not None:
            spectrophot_blksize = len(self.tds.spectrophotometric_data)
            if spectrophot_blksize > 0:
                offset = phot_blksize + spec_blksize
                self.spectrophot_vals = vals[offset:]
        # self.phot_vals = vals[:n]
        # self.spec_vals = vals[n:]
        # self.spectrophot_vals = vals[]

    def set_sn_pars(self, z, x0, x1, c, tmax, spec_mjd=None):
        """
        """
        # SN local parameters
        self.tds.sn_data.nt['z'] = z
        self.tds.sn_data.nt['x0'] = x0
        self.tds.sn_data.nt['x1'] = x1
        self.tds.sn_data.nt['c'] = c
        self.tds.sn_data.nt['tmax'] = tmax

        # redshift
        if self.tds.lc_data:
            self.tds.lc_data.z[:] = z
        if self.tds.spec_data:
            self.tds.spec_data.z[:] = z
        if self.tds.spectrophotometric_data:
            self.tds.spectrophotometric_data.z[:] = z

        # mjd
        if self.tds.lc_data:
            self.tds.lc_data.nt['mjd'] = self.lc_mjd + tmax

        if spec_mjd is None:
            spec_mjd = tmax
        if self.tds.spec_data:
            self.tds.spec_data.nt['mjd'] = spec_mjd
        if self.tds.spectrophotometric_data:
            self.tds.spectrophotometric_data.nt['mjd'] = spec_mjd

    def load_sn_pars(self, sn, tds, pars=None, spec_mjd=None):
        """
        """
        idx = tds.sn_data.sn == sn
        assert idx.sum() == 1
        s = tds.sn_data.nt[idx]
        if pars is None:
            self.set_sn_pars(s.z, s.x0, s.x1, s.col, s.tmax, spec_mjd=spec_mjd)
        else:
            sn_index = tds.sn_data.sn_map[sn]
            self.set_sn_pars(s.z,
                             pars['X0'].full[sn_index],
                             pars['X1'].full[sn_index],
                             pars['c'].full[sn_index],
                             pars['tmax'].full[sn_index],
                             spec_mjd=spec_mjd)

        # load the light curve zp
        # todo: simplify this ! A tds should be able to
        # retrieve the zp of a given SN light curve.
        if self.bands is not None:
            for band in self.bands:
                idx = (tds.lc_data.sn == sn) & (tds.lc_data.band == band)
                local_idx = self.tds.lc_data.band == band
                # if the band is not in the tds
                # we set the zp to, say, 15.
                if idx.sum() == 0:
                    logging.warning(f'{band}: no data for sn#{sn}')
                    self.tds.lc_data.zp[local_idx] = 15.
                else:
                    zp = tds.lc_data.zp[idx]
                    assert np.all(zp == zp[0])
                    sel = self.tds.lc_data.band == band
                    self.tds.lc_data.zp[local_idx] = zp[0]
            self.tds.compute_photometric_norm_factors()

        # evaluate the model
        # TODO: we shouldn't have to re-instantiate the model...
        self.model = self.model_type(self.tds,
                        init_from_salt2_file=self.init_from_salt2_file)
        self.model.init_from_training_dataset()
        self(self.model.pars.free)

    def plot_lc(self, band, ax=None, **kwargs):
        """
        """
        if not ax:
            ax = plt.gca()
        idx = self.tds.lc_data.band == band
        if idx.sum() == 0:
            return
        ax.plot(self.tds.lc_data.mjd[idx], self.phot_vals[idx], **kwargs)

    def plot_spectrum(self, ax=None, phot=False, spec=0, **kwargs):
        """
        """
        spec_data, vals = None, None
        if not phot:
            spec_data = self.tds.spec_data
            vals = self.spec_vals
        else:
            spec_data = self.tds.spectrophotometric_data
            vals = self.spectrophot_vals
        assert spec_data is not None and vals is not None

        idx = spec_data.spec == spec
        if idx.sum() == 0:
            return
        ax = kwargs.get('ax', None)
        if not ax:
            ax = plt.gca()
        ax.plot(spec_data.wavelength[idx],
                vals[idx], **kwargs)



import numpy as np
import pylab as pl
from nacl import dataset
from nacl.dataset import TrainingDataset
from nacl.models.salt import SALT2Like
from nacl.models.constraints import SALT2LikeConstraints
from nacl.models.regularizations import NaClSplineRegularization
from nacl import minimize
import helpers

jla_tds = None

def load():
    global jla_tds
    if jla_tds is None:
        jla_tds = dataset.read_hdf('jla.hd5')
    return jla_tds

def fit(tds):
    model = SALT2Like(tds, init_from_salt2_file='salt2.npz')
    model.init_from_training_dataset()

    #  light curve fit only. Invalidate the spectral data
    valid = tds.spec_data.valid.copy()
    tds.spec_data.valid[:] = 0

    # eliminate all measurements before -20 and after +50
    idx = (tds.lc_data.mjd - model.pars['tmax'].full[tds.lc_data.sn_index])>50
    tds.lc_data.valid[idx] = 0
    idx = (tds.lc_data.mjd - model.pars['tmax'].full[tds.lc_data.sn_index])<-25
    tds.lc_data.valid[idx] = 0

    # remove u-band data
    tr = tds.get_all_transmissions()
    for k in tr:
        wl = tr[k].mean()
        if wl < 3900:
            idx = tds.lc_data.band == k
            print(f'killed {idx.sum()} measurements in {k}')
            tds.lc_data.nt.valid[idx] = 0

    # keep only the supernovae with points at least 5 days before max.
    # we need to be conservative, because tmax may vary during the fit
    idx = (tds.lc_data.mjd - model.pars['tmax'].full[tds.lc_data.sn_index])<-5
    sne = [tds.sn_data.sn_map[sni] for sni in \
           np.where((np.bincount(tds.lc_data.sn_index, idx) == 0))[0]]
    tds.kill_sne(sne)

    # remove the supernovae with strange colors
    idx = np.abs(tds.sn_data.col) > 1.
    sne = tds.sn_data.sn[idx]
    tds.kill_sne(sne)
    tds.compress()

    # re-instantiate the model
    model = SALT2Like(tds, init_from_salt2_file='salt2.npz')
    model.init_from_training_dataset()

    model.pars.fix()
    model.pars['X0'].release()
    model.pars['X1'].release()
    model.pars['col'].release()
    model.pars['tmax'].release()

    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres)
    minz = minimize.Minimizer(chi2)

    return minz






"""Pure python reimplementation of the original salt2 model
"""


import logging
import numpy as np
import scipy.sparse
import sncosmo

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt

from bbf import bspline, SNFilterSet, SNMagSys
from saltworks import FitParameters

from .colorlaw import ColorLaw
from .lightcurves import LightcurveEvalUnit
from .spectra import CompressedSpectrumEvalUnit, CompressedSpectroPhotoEvalUnit #  SpectrumRecalibrationPolynomials



class SALT2Like(object):
    r"""A re-implementation of the SALT2 model

    SALT2 is an empirical SN spectrophotometric model. This class provides a
    pure python re-implementation of SALT2 with a number of improvements.

    The SALT2 parametrization is defined as follows. In the SN restframe the
    absolute spectral density :math:`S(\lambda)` is:

    .. math::
        S(\lambda, \mathrm{p}) = X_0 \times \left[M_0(\lambda, \mathrm{p}) + X_1
           \times M_1(\lambda, \mathrm{p})\right]\ 10^{0.4 c CL(\lambda)}

    where :math:`\mathrm{p}` is the SN phase, i.e. the restframe time since SN
    peak luminosity:

    .. math::
        \mathrm{ph} = \frac{t_{MJD} - t_{max}}{1 + z}

    :math:`M_0`, :math:`M_1` are global surfaces describing the spectral
    evolution of the average SN and its principal variation, and
    :math:`CL(\lambda)` is a global "extinction correction" describing the color
    diversity of the SNIa family.

    :math:`(X_0, X_1, c)` are SN-dependent. :math:`X_0` is the amplitude of the
    "SN-frame B-band lightcurve" as inferred from the observer-frame fluxes.
    :math:`X_1` is the coordinate of the SN along :math:`M_1`. In practice, it
    quantifies the variations of the light curve width. :math:`c` is the SN
    color -- or more exactly, the SN color excess with respect to the average SN
    color.

    """

    def __init__(self, training_dataset,
                 phase_range=(-20., 50.), wl_range=(2000., 9000.),
                 basis_knots=[127, 20],
                 basis_filter_knots=900,
                 wl_grid=None, phase_grid=None,
                 spectrum_recal_degree=3,
                 normalization_band_name='swope2::b'):
        """Constructor

          - instantiate the bases (phase, wavelength, filter)
          - compute the Gram
          - compute the lambda_eff matrix
          - instantiate a color law and evaluate it on the lambda_eff's
          - organizes the data into spectral/photometric computing units

        Parameters
        ----------
        training_dataset : nacl.dataset.TrainingDataset
            Training dataset.
        phase_range : 2-tuple
            Nominal phase range of the model.
        wl_range : 2-tuple
            Nominal wavelength range of the model.
        basis_knots : 2-list
            Number of wavelength and photometric knot.
        basis_filter_knots : int
            Number of knot for the filter basis
        spectrum_recal_degree : int
            Degree of the spectrum recalibration polynomials
        normalization_band_name : str
            Use this band to normalize the model.
        """
        self.training_dataset = training_dataset
        self.lc_data = training_dataset.lc_data
        self.phase_range = phase_range
        self.wl_range = wl_range
        self.n_wl, self.n_ph = basis_knots[0],  basis_knots[1]
        self.basis_knots = basis_knots
        self.basis_filter_knots = basis_filter_knots
        self.spectrum_recal_degree = spectrum_recal_degree
        self.delta_phase = 0. # check: do not remember what this is
        # self.filterpath = self.training_dataset.filterpath
        if self.training_dataset.lc_data is not None:
            self.bands = list(self.training_dataset.transmissions.keys())
        self.normalization_band_name = normalization_band_name
        # self.calib_variance = calib_variance

        self.timing = []

        self.basis, self.gram, self.G2, self.L_eff = None, None, None, None
        self.val, self.jacobian_val, self.jacobian_i, self.jacobian_j = None, None, None, None
        self.polynome_color_law, self.jacobian_color_law = None, None

        # Filter database
        # TODO: we should be able to optimize this a little
        # (i.e. adapt the filter bases, to lower the size of the grams)
        self.filter_basis = bspline.BSpline(np.linspace(self.wl_range[0],
                                                        self.wl_range[1],
                                                        basis_filter_knots), order=4)
        self.filter_db = SNFilterSet(basis=self.filter_basis)
        self.filter_db.insert('swope2::b', z=0.)
        self.color_law = ColorLaw()

        n_lc_meas, n_spec_meas, n_spectrophot_meas = \
            self.training_dataset.nb_meas(split_by_type=1)

        # TODO: where is this used ?
        # TODO: we may replace this by n_lc_meas > 0
        if self.training_dataset.lc_data is not None:
            tds = self.training_dataset
            ii = self.training_dataset.i_lc_first
            self.lambda_c = tds.lc_data.wavelength[ii] / (1+tds.lc_data.z[ii])
            self.lambda_c_red = self.color_law.reduce(self.lambda_c)

        # initialize degree of the spectral recalibration polynomial
        # self._init_spectral_polynomial()

        # initialize model basis and model parameter vector
        # model component: bases, color law, and parameter vector
        self._init_bases(wl_grid, phase_grid,
                         wl_range, phase_range, basis_knots)

        # OK. For now, the color scatter and the calibration scatter
        # are explicitely instantiated in the model constructor
        #        self.error_snake_model = self._instantiate_error_model(error_snake_model_type)
        #        self.calib_error_model = self._instantiate_error_model(calib_error_model_type)
        #        self.color_scatter_model = self._instantiate_error_model(color_scatter_model_type)

        # finally, it is better to have it here.
        # NOTE: if there are spectra, the model must be evaluated for them.
        # even if they are marked as invalid
        #self.recal_func = None
        #if self.training_dataset.spec_data is not None:
        #    self.recal_func = \
        #        SpectrumRecalibrationPolynomials(self.training_dataset, self,
        #                                         self.recalibration_degree)

        # initialize the global model parameters (M0, M1, CL)
        # self.pars = self.init_pars()

        # some of the classes gravitating around the model
        # the constraints for example, sometimes need to have
        # access to the index of some paramters in the full parameter
        # vector (all parameters released)
        # self.all_pars_released = self.pars.copy()
        # self.all_pars_released.release()

        # initialize the SN specific parameters (X0, X1, col, tmax)
        # if init_from_training_dataset:
        #     self.init_from_training_dataset()

        # finalize the initialization of the error models 
        # which need a fully operational parameter vector 
#        if self.calib_error_model is not None:
#            self.calib_error_model.finalize()

        # we use a default model normalization
        # this is what guarantees that the SALT2.4 absolute mag
        # is -19.5 in the SWOPE::B band. Of course, this default
        # normalization may be changed explicitely, by calling
        # self.renorm(band_name=, Mb=, magsys='AB')
        self.norm = self.normalization(pars=None, band_name=normalization_band_name,
                                       default_norm=1.01907246e-12)

        # grams
        self.init_grams_and_cc_grid()

        # and finally, prepare the computing units
        self.queue = self._init_eval_units()

        # clear the cache
        # self.clear_cache()

        # if we need to run queue units separately,
        # could be useful to disable caching of the results
        # self.disable_cache = False

    # def clone(self, training_dataset):
    #     """return a clone (same basis, same color law) for a different tds
    #     """
    #     ret = SALT2Like(training_dataset,
    #                     phase_range=self.phase_range,
    #                     wl_range=self.wl_range,
    #                     basis_knots=self.basis_knots,
    #                     basis_filter_knots=self.basis_filter_knots,
    #                     spectrum_recal_degree=self.spectrum_recal_degree,
    #                     normalization_band_name=self.normalization_band_name,
    #                     error_snake_model=self.error_snake_model,
    #                     calib_error_model=self.calib_error_model,
    #                     color_scatter_model=self.color_scatter_model)
    #                      calib_variance=self.calib_variance)
    #     for block_name in ['M0', 'M1', 'CL']:
    #         ret.pars[block_name].full[:] = self.pars[block_name].full[:]
    #     ret.norm = ret.normalization(band_name=ret.normalization_band_name)

    #     return ret

    def get_gram_dot_filter(self, band_name=None):
        """Used by the integral constraints
        """
        key = band_name if band_name is not None else self.normalization_band_name
        coeffs, _ = self.filter_db[key]
        return self.gram.dot(coeffs)

    def reduce(self, wl):
        return self.color_law.reduce(wl)

    # def _instantiate_error_model(self, error_model_type):
    #     if error_model_type is None:
    #         return None
    #     if type(error_model_type) == tuple:
    #         cls, args = error_model_type
    #         return cls(self, **args)
    #     if type(error_model_type) == type:
    #         return error_model_type(self)
    #     raise TypeError('arg should either be: None|(model_class, (args))|(model_class)')

    def init_grams_and_cc_grid(self):
        r"""Compute the :math:`\Lambda^{\mathrm{eff}}_{\ell q}` matrix (see definition above) on which the color
        law is evaluated.

        .. math::
             \bar{\lambda}_{\ell q} = \frac{\int \lambda^2 B_\ell(\lambda) B_q(\lambda)
             d\lambda}{\int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda}

        Compute de grammian of order one, :math:`G` and two :math:`G2` of the model :

        .. math::
            G = \int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda \\
            G2 = \int \lambda^2 B_\ell(\lambda) B_q(\lambda) d\lambda


        .. note::
             this method will be moved to the main model class (since we may work
             with one single :math:`\Lambda^{\mathrm{eff}}` matrix in a near future).
        """
        self.gram = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=1)
        self.G2 = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=2)

        gram = self.gram.tocoo()
        gram2 = self.G2.tocoo()
        assert(~np.any(gram.row-gram2.row) and ~np.any(gram.col-gram2.col))
        l_eff = gram2.data / gram.data
        self.L_eff = scipy.sparse.coo_matrix((l_eff, (gram.row, gram.col)), shape=gram.shape)

    def get_struct(self):
        """return the structure of the fit parameter vector

        In practice, the fit parameter vector is larger than just the model
        parameters: it also contains the parameters of the error models (error
        snake, color scatter, calibration, see e.g. variancemodels.py). The
        instantiation of the final fit parameter vector cannot therefore be
        performed by the model itself. What the model can do, is to return the
        description of the fit parameter blocks it knows about.

        """
        nb_sne = self.training_dataset.nb_sne()
        # nb_spectra = self.training_dataset.nb_spectra()
        # nb_passbands = len(self.bands)
        # spec_recalibration_npars = self.recalibration_degree + 1
        nb_lightcurves = self.training_dataset.nb_lcs(valid_only=False)
        d = [('X0',   nb_sne),
             ('X1',   nb_sne),
             ('c',  nb_sne),
             ('tmax', nb_sne),
             ('M0',   self.basis.bx.nj * self.basis.by.nj),
             ('M1',   self.basis.bx.nj * self.basis.by.nj),
             ('CL',  4)]
        if hasattr(self, 'recal_func'):
            d.extend(self.recal_func.get_struct())
            #        if self.training_dataset.spec_data is not None:
            #            d.append(('SpectrumRecalibration', spec_recalibration_npars.sum()))
        return d

    def init_pars(self, pars=None, model_name='salt2.4', version='2.4'):
        """instantiate a fit parameter vector

        Returns
        -------
        fp : nacl.lib.fitparameters.FitParameters
            Model parameters.
        """
        if pars is None:
            pars = FitParameters(self.get_struct())

        self.init_model_pars(pars, model_name=model_name, version=version)
        self.init_from_training_dataset(pars)

        if 'SpectrumRecalibration' in pars._struct:
            self.recal_func.init_pars(pars)

        return pars

    def init_from_training_dataset(self, pars):
        """load initial sn parameters from the training dataset
        """
        sn_data = self.training_dataset.sn_data
        pars['X0'].full[sn_data.sn_index] = sn_data.x0
        pars['X1'].full[sn_data.sn_index] = sn_data.x1
        pars['c'].full[sn_data.sn_index] = sn_data.col
        pars['tmax'].full[sn_data.sn_index] = sn_data.tmax

    def init_model_pars(self, pars, model_name='salt2', version='2.4'):
        """Load & adapt the SALT2.4 global surfaces and color law from sncosmo.

        Load the coefficients of the SALT2.4 surfaces and color law. The
        definition of the model spline bases differs from the original
        definition of the SALT2.4 bases; so, we reproject the original SALT2
        surfaces on the model basis.

        Parameters
        ----------
        salt2_filename : str

        """
        salt2_source = sncosmo.get_source('salt2', version='2.4')

        phase_grid = np.linspace(self.phase_range[0], self.phase_range[1], self.n_ph)
        wl_grid = np.linspace(self.wl_range[0], self.wl_range[1], self.n_wl)
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)

        phase_salt = salt2_source._phase
        wl_salt = salt2_source._wave

        w, p = np.meshgrid(wl_salt, phase_salt)

        sncosmo_scale = salt2_source._SCALE_FACTOR
        salt2_m0 = salt2_source._model['M0'](phase_salt, wl_salt).ravel() / sncosmo_scale
        salt2_m1 = salt2_source._model['M1'](phase_salt, wl_salt).ravel() / sncosmo_scale
        salt2_cl = np.array(salt2_source._colorlaw_coeffs)[::-1]

        jac = self.basis.eval(w.ravel(), p.ravel()).tocsr()
        factor = cholesky_AAt(jac.T, beta=1.E-6)

        pars['M0'].full[:] = factor(jac.T * salt2_m0)
        pars['M1'].full[:] = factor(jac.T * salt2_m1)
        pars['CL'].full[:] = salt2_cl

    # def _init_spectral_polynomial(self):
    #     """instantiate the recalibration polynomials

    #     if a degree is specified, all spectra have the same degree. Since the
    #     degree of the polynomial must not exceed the number of Light curves :
    #     this degree is large odd number inferior to the numbers of light curves
    #     for this SN.

    #     """
    #     # note: the model is evaluated on the full dataset.
    #     # not just the valid data.
    #     nb_spectra = self.training_dataset.nb_spectra(valid_only=False)
    #     self.recalibration_degree = np.full(nb_spectra, self.spectrum_recal_degree)

    def _init_bases(self, wl_grid=None, phase_grid=None,
                    wl_range=(2000., 9000.),
                    phase_range=(-20., 50.), basis_knots=(127,20)):
        """
        Instantiate model bases
        """
        if wl_grid is None:
            assert wl_range is not None and basis_knots is not None
            logging.info('default regular grid in wavelength')
            wl_grid = np.linspace(self.wl_range[0], self.wl_range[1],
                                  self.n_wl)
        else:
            logging.info('user provided grid in wavelength')
        if phase_grid is None:
            assert phase_range is not None and basis_knots is not None
            logging.info('default regular grid in phase')
            phase_grid = np.hstack([np.linspace(self.phase_range[0],
                                                self.phase_range[1],
                                                self.n_ph)])
        else:
            logging.info('user provided grid in phase')
        self.basis = bspline.BSpline2D(wl_grid, phase_grid,
                                       x_order=4, y_order=4)

    #    @property
    #    def noise_models(self):
    #        """return a list of all the noise models contained by the model
    #        """
    #        ret = [self.calib_error_model,
    #               self.color_scatter_model,
    #               self.error_snake_model]
    #        return list(filter(lambda x: x is not None, ret))

    def normalization(self, pars=None, band_name='swope2::b', Mb=-19.5, magsys='AB',
                      default_norm=1.01907246e-12):
        """model normalization

        The SALT2Like normalization is set during training by the constraint
        on the integral of M0 at phase zero.

        Therefore, by default, we do not renormalize the model during
        evaluation.
        """
        default_norm = default_norm if default_norm is not None else 1.01907246e-12
        tq, filter_basis = self.filter_db[band_name]

        # AB flux in the specified band
        ms = SNMagSys(self.filter_db, magsys)
        # zp = ms.ZeroPoint(self.filter_db.transmission_db[band_name])
        zp = ms.get_zp(band_name)
        self.int_ab_spec = 10**(0.4 * zp)

        # normalization quantities
        self.flux_at_10pc = np.power(10., -0.4 * (Mb-zp))
        self.flux_at_10Mpc = np.power(10., -0.4 * (Mb+30.-zp))

        if pars is None:
            return default_norm

        phase_eval = self.basis.by.eval(np.array([0. + self.delta_phase])).tocsr()
        gram = get_gram(z=0., model_basis=self.basis.bx,
                        filter_basis=filter_basis, lambda_power=1)
        surface_0 = pars['M0'].full.reshape(len(self.basis.by), -1)
        # evaluate the integral of the model in the specified band
        self.int_M0_phase_0 = phase_eval.dot(surface_0.dot(gram.dot(tq)))

        return self.flux_at_10Mpc / self.int_M0_phase_0

    def renorm(self, pars, band_name='swope2::b', Mb=-19.5, magsys='AB'):
        """Adjust the model normalization
        """
        # explicitely recompute a normalization
        self.norm = self.normalization(pars, band_name, Mb, magsys,
                                       default_norm=None)

    def _init_eval_units(self):
        """Initialize the photometric and spectroscopic eval units

        The actual evaluation of the model is deferred to specialized
        sub-units, which are called one after the other. These units
        are stored in a queue.

        """
        # n_lc_meas, n_spec_meas, n_spectrophot_meas = \
        #     self.training_dataset.nb_meas(valid=True, split_by_type=True)
        queue = []
        if self.training_dataset.lc_data is not None:
            logging.info('initializing lightcurve eval unit')
            queue.append(LightcurveEvalUnit(self))
        if self.training_dataset.spec_data is not None:
            logging.info('initializing spectrum eval unit')
            queue.append(CompressedSpectrumEvalUnit(self, spec_recal_degree=self.spectrum_recal_degree))
        if self.training_dataset.spectrophotometric_data is not None:
            logging.info('initializing spectrophotometric eval unit')
            queue.append(CompressedSpectroPhotoEvalUnit(self))
        return queue

    # def update_computing_units(self, ilc=None, spec=False):
    #     """
    #     Recreate the model queue as a function of the wanted LCs and spectra.

    #     Parameters
    #     -------
    #     ilc : None or int
    #         If None all light curve are evaluated, else only teh desired one.
    #     spec : bool
    #         Whether Spectra need to be evaluated.

    #     Returns
    #     -------
    #     queue : list
    #         The new queue
    #     """
    #     queue = []
    #     if ilc is not None:
    #         for lc in [self.training_dataset.lcs[ilc]]:
    #             queue.append(LightcurveEvalUnitTz(lc, self))
    #     if spec:
    #         queue.append(SpectrumEvalUnitFast(self.training_dataset, self))
    #     return queue

    # def clear_cache(self):
    #     """
    #     Clear model evaluation, derivatives and timing.
    #     """
    #     self.val = []
    #     self.jacobian_i = []
    #     self.jacobian_j = []
    #     self.jacobian_val = []
    #     self.timing = []

    def precompute_color_law(self, cl_pars, jac=False):
        """
        Precompute the color law for the photometric data.

        Parameters
        -------
        cl_pars : None or numpy.array
            Color law parameters
        jac : bool
            If derivatives are needed.
        """
        polynome_color_law, jacobian_color_law = \
            self.color_law(self.L_eff.data, cl_pars,
                           jac=jac)  # return_jacobian_as_coo_matrix=False)
        self.polynome_color_law = \
            scipy.sparse.csr_matrix((polynome_color_law,
                                     (self.L_eff.row, self.L_eff.col)),
                                    shape=self.L_eff.shape)
        if jac:
            self.jacobian_color_law = []
            _, n = jacobian_color_law.shape
            for i in range(n):
                jacobian_cl = \
                    scipy.sparse.csr_matrix((jacobian_color_law[:, i],
                                             (self.L_eff.row, self.L_eff.col)), shape=self.L_eff.shape)
                self.jacobian_color_law.append(jacobian_cl)

    def get_restframe_phases(self, pars, data):
        """
        Given a chunk of data, return the restframe phases

        Returns
        -------
        numpy.array
            data SN-restframe phase
        """
        sn_index = data.sn_index
        tmax = pars['tmax'].full[sn_index]
        return (data.mjd - tmax) / (1.+data.z)

    def __call__(self, pars, jac=False): # , plotting=False, ilc_plot=None, spec_plot=False):
        """Evaluate the model for the parameter set p

        In practice, loop over the queue and run each eval unit.
        Assemble the results into one single vector and (optionally)
        one single jacobian matrix.

        Parameters
        -------
        p : numpy.array
            Vector containing the free parameters only.
        jac : bool
            Whether it return the jacobian matrix.

        Returns
        -------
        val : numpy.array
            model results
        jacobian : scipy.sparse.csr_matrix
            jacobian matrix (if jac is true)
        """
        # pre-evaluate the color law (shared between all light curves)
        # cl_pars = pars['CL'].full
        # self.precompute_color_law(cl_pars, jac=jac)

        # # precompute the values of the phase basis on the full dataset
        # # significantly faster that way
        # if self.training_dataset.lc_data is not None:
        #     restframe_phases = self.get_restframe_phases(pars, self.lc_data)
        #     self.phase_eval = self.basis.by.eval(restframe_phases + self.delta_phase).tocsr()
        #     if jac:
        #         self.dphase_dtmax = self.basis.by.deriv(restframe_phases + self.delta_phase).tocsr()
        #     else:
        #         self.dphase_dtmax = None

        # self.clear_cache()

        # now, we are ready to loop over the eval units
        res = [q(pars, jac) for q in self.queue]


        if not jac:
            model_val = np.add.reduce([r for r in res])
            return model_val

        model_val = np.add.reduce([r[0] for r in res])

        rows = np.hstack([r[1].row for r in res])
        cols = np.hstack([r[1].col for r in res])
        vals = np.hstack([r[1].data for r in res])

        idx = cols >= 0
        JJ = scipy.sparse.coo_matrix((vals[idx], (rows[idx], cols[idx])),
                                     shape=res[0][1].shape)
        # JJ = scipy.sparse.dok_matrix(res[0][1].shape)
#        for r in res:
#            JJ += r[1]
#
        return model_val, JJ

        # n_data = self.training_dataset.nb_meas(valid_only=False)

        # logging.debug('computing derivatives: hstack...')
        # n = len(pars.free)
        # i = np.hstack(self.jacobian_i)
        # j = np.hstack(self.jacobian_j)
        # v = np.hstack(self.jacobian_val)
        # logging.debug('building coo_matrix...')
        # idx = j >= 0  # self.pars.indexof(j) >= 0
        # jacobian = scipy.sparse.coo_matrix((v[idx], (i[idx], j[idx])), shape=(n_data, n))
        # logging.debug('ok, done.')
        # return val, jacobian


    @property
    def y(self):
        return self.training_dataset.get_all_fluxes()

    @property
    def yerr(self):
        return self.training_dataset.get_all_fluxerr()

    @property
    def bads(self):
        return self.training_dataset.get_bads()


_G, _G2 = {}, {}


def get_gram(z, model_basis, filter_basis, lambda_power=1):
    """
    Calculate the grammian of to spline basis.

    The grammian :math:`G` of order :math:`N`, to basis :math:`B_0` (define on wavelength,
    in our case surfaces wavelength basis) and
    :math:`B_1` (define on SN-restframe wavelength, filter basis) is defined as :

    .. math::
        G = \\int \\lambda^N B_0(\\lambda) B_q(\\lambda (1+z)) d\\lambda

    Parameters
    -------
    z : numpy.array
        Vector containing the data redshift.
    model_basis : nacl.lib.bspline.BSpline
        Wavelength basis.
    filter_basis : nacl.lib.bspline.BSpline
        Filter basis defined common to all the bands.
    lambda_power : int
        Grammian order.

    Returns
    -------
    gram : scipy.sparse.csc_matrix
        Grammian
    """
    global _G
    global _G2
    key = (z, model_basis, filter_basis, lambda_power)

    gram = _G.get(key, None)
    if gram is None:
        gram = bspline.lgram(model_basis, filter_basis, z=z, lambda_power=lambda_power).tocsc()
        _G[key] = gram

    return gram




# class SALT2Eval:
#     """Evaluate the model, for one single SN
#     """
#     def __init__(self, model, bands=['SWOPE::B'],
#                  n_spectra=1, n_phot_spectra=1):
#         """
#         """
#         self.orig_model = model
#         self.phase_range = model.phase_range
#         self.wl_range = model.wl_range
#         self.tds = SimTrainingDataset(bands,
#                                       n_spectra=n_spectra,
#                                       n_phot_spectra=n_phot_spectra)
#         # it is essential that the model is an exact duplicate
#         # of the original model - just 1 SN instead of many
#         self.model = model.clone(self.tds)

#     # def clone_model(self):
#     #     """instantiate a duplicate of the original model

#     #     .. note :: a model should be able to clone itself - would be cleaner
#     #     """
#     #     model = self.orig_model
#     #     ret = \
#     #         SALT2Like(self.tds,
#     #                   phase_range=model.phase_range,
#     #                   wl_range=model.wl_range,
#     #                   basis_knots=model.basis_knots,
#     #                   basis_filter_knots=model.basis_filter_knots,
#     #                   spectrum_recal_degree=model.spectrum_recal_degree,
#     #                   normalization_band_name=model.normalization_band_name,
#     #                   calib_variance=model.calib_variance)
#     #     return ret

#     def set(self, **kwargs):
#         """initialize the sn parameters

#         This function updates the dataset, and then the model
#         parameters.
#         """
#         # sn parameters
#         if 'z' in kwargs:
#             z = kwargs.get('z')
#             self.tds.sn_data.nt['z'] = z
#             if self.tds.lc_data:
#                 self.tds.lc_data.z[:] = z
#             if self.tds.spec_data:
#                 self.tds.spec_data.z[:] = z
#             if self.tds.spectrophotometric_data:
#                 self.tds.spectrophotometric_data.z[:] = z
#         if 'x0' in kwargs:
#             self.tds.sn_data.nt['x0'] = kwargs['x0']
#         if 'x1' in kwargs:
#             self.tds.sn_data.nt['x1'] = kwargs['x1']
#         if 'c' in kwargs:
#             self.tds.sn_data.nt['c'] = kwargs['c']
#         if 'tmax' in kwargs:
#             tmax = kwargs['tmax']
#             self.tds.sn_data.nt['tmax'] = kwargs['tmax']
#             if self.tds.lc_data:
#                 self.tds.lc_data.nt['mjd'] = self.tds.mjd +tmax
#         if 'ebv' in kwargs:
#             self.tds.sn_data.nt['ebv'] = kwargs['ebv']

#         if 'spec_mjd' in kwargs:
#             if self.tds.spec_data:
#                 self.tds.spec_data.nt['mjd'] = kwargs['spec_mjd']
#             if self.tds.spectrophotometric_data:
#                 self.tds.spectrophotometric_data.nt['mjd'] = kwargs['spec_mjd']

#         # light curve zero points and magsys
#         if 'zp' in kwargs:
#             zps = kwargs['zp']
#             for band_name in zps:
#                 idx = self.tds.lc_data.band == band_name
#                 self.tds.lc_data.nt['zp'] = zps[band_name]
#         if 'magsys' in kwargs:
#             magsys = kwargs['magsys']
#             if type(magsys) is str:
#                 self.tds.lc_data.nt['magsys'][:] = magsys
#             for band_name in magsys:
#                 idx = self.tds.lc_data.magsys == band_name
#                 self.tds.lc_data.nt['magsys'] = magsys[band_name]

#         # recompute the photometric norm factors
#         self.tds.compute_photometric_norm_factors()
#         # propagate this into the model parameters
#         self.model = self.orig_model.clone(self.tds)
#         self.model.init_from_training_dataset()

#     def set_from_tds(self, sn, tds):
#         """load the sn parameters from a training dataset
#         """
#         sn_pars = tds.get_sn_pars(sn)
#         if self.tds.lc_data:
#             sn_pars['zp'] = tds.get_zp(sn)
#             sn_pars['magsys'] = tds.get_magsys(sn)
#         self.set(**sn_pars)
#         # self.set(z=sn_pars['z'], x0=sn_pars['x0'],
#         #          x1=sn_pars['x1'], col=sn_pars['c'],
#         #          tmax=sn_pars['tmax'], ebv=sn_pars['ebv'],
#         #          zp=zp, magsys=magsys)

#     def update_global_pars(self, pars):
#         """
#         """
#         for block_name in ['M0', 'M1', 'CL']:
#             self.orig_model.pars[block_name].full[:] = pars[block_name].full[:]
#             self.model.pars[block_name].full[:] = pars[block_name].full[:]

#     def __call__(self):
#         """evaluate the model and update the dataset with the model values

#         This is the simplest way to evaluate the model.
#         No need to clone the dataset or anything.
#         """
#         v = self.model(self.model.pars.free)
#         self.tds.update_flux(v)

#     def bandflux(self, bands, mjds, zp, zpsys):
#         """evaluate the model for the requested bands and at the requested mjds

#         This one in a little more costly, as we need to instantiate
#         a new training dataset and re-instantiate a model.
#         """
#         pass

#     def bandmag(self, bands, mjd, magsys):
#         """return mags for the requested bands and at the requested mjds

#         This is also a little more costly, as we need to instantiate
#         a new training dataset and re-instantiate a model.
#         """
#         pass

#     def spectrum(self, mjd):
#         """evaluate and return a spectrophotometric spectrum
#         """
#         pass

#     def photspectrum(self, mjd):
#         """evaluate and return a spectrophotometric spectrum
#         """
#         pass

#     def color_law(self, col=None):
#         pass


# # def main():
# #     model = ...
# #     m = SALT2LikeEval(model, bands=['SWOPE::B'])

# #     # evaluate the model on all the data for this SN
# #     # with the default parameters
# #     m.eval(sn=sn)

# #     # evaluate the model on all the data for this SN
# #     # with alternative parameters
# #     m.eval(sn=sn, sn_pars=sn_pars, model_pars=model_pars)



    # def init_from_salt2_old(self, salt2_filename, stick_to_original_model=False):
    #     """Load & adapt the SALT2.4 global surfaces and color law.

    #     Load the coefficients of the SALT2.4 surfaces and color law. The
    #     definition of the model spline bases differs from the original
    #     definition of the SALT2.4 bases; so, we reproject the original SALT2
    #     surfaces on the model basis.

    #     Parameters
    #     ----------
    #     salt2_filename : str
    #         the classical salt2.npz filename containing the definition of the
    #         SALT2.4 bases, M0, M1 surfaces and color_law.
    #     """
    #     f = np.load(salt2_filename)
    #     phase_grid = f['phase_grid']
    #     wl_grid = f['wl_grid']
    #     basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)

    #     if stick_to_original_model:
    #         # don't remember why we have this
    #         self.delta_phase = +0.7
    #         self.basis = basis
    #         self.pars = self.init_pars()
    #         self.pars['M0'].full[:] = f['M0'].T.ravel()
    #         self.pars['M1'].full[:] = f['M1'].T.ravel()
    #         self.pars['CL'].full[:] = f['CL_pars'][0:4]
    #         return

    #     # TODO: replace this with
    #     # xx = self.basis.bx.grid
    #     # yy = self.basis.by.grid
    #     xx = np.linspace(wl_grid[0], wl_grid[-1], basis.bx.nj)
    #     yy = np.linspace(phase_grid[0], phase_grid[-1], basis.by.nj)
    #     x,y = np.meshgrid(xx,yy)
    #     x,y = x.ravel(), y.ravel()
    #     jac = self.basis.eval(x,y).tocsr()
    #     factor = cholesky_AAt(jac.T, beta=1.E-6)

    #     # and initialize the parameters
    #     self.pars = self.init_pars()
    #     self.pars['M0'].full[:] = factor(jac.T * f['M0'])
    #     self.pars['M1'].full[:] = factor(jac.T * f['M1'])
    #     self.pars['CL'].full[:] = f['CL_pars'][0:4]



import numpy as np
import pylab as pl
from nacl.minimize import WeightedResiduals
import helpers

tds, model = helpers.generate_dataset(2, seed=42)


def check_wres_grad(wres, p, dx=1.E-5):
    v0, jacobian, _ = wres(p, jac=True)
    p0 = p.copy()

    df = []
    for i in range(len(p)):
        p0[i] += dx
        vp = wres(p0, jac=False)
        df.append((vp-v0) / dx)
        p0[i] -= dx
    return np.array(jacobian.todense()), np.vstack(df).T


def test_weighted_residual_derivatives():
    """
    """
    wres = WeightedResiduals(model)
    jac, num_der = check_wres_grad(wres, p=model.pars.free)

    return np.array(jac), num_der



#!/usr/bin/env python3

import numpy as np
import pylab as pl

from nacl.models.toy2d import lightcurves
from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer
from matplotlib.gridspec import GridSpec
from nacl.lib.fitparameters import FitParameters

from nacl.models.helpers import check_grad, check_deriv_old

from scipy import sparse




def main(release=['x0', 'tmax'], error_pedestal=0.05):
    # generate dataset
    gen = lightcurves.ToyModelGenerator(nsn = 10, nbands=1, npts1=20, npts2=30, yerr=0.01, error_pedestal=error_pedestal)
    dp = gen.generate()

    # build a model and a log-likelihood
    model = lightcurves.Model(dp)
    reg = lightcurves.Regularization(block_name='theta')
    cons = lightcurves.cons(model, mu=1.E6)
    snake = lightcurves.SimplePedestalModel(model)
    ll = LogLikelihood(model, variance_model=snake, reg=[reg], cons=[cons], data=dp)

    # fit initialization
    for block_name in ['x0', 'tmax', 'stretch']:
        ll.pars[block_name] = dp.sample[block_name]
        ll.pars[block_name].fix()

    for block_name in release:
        ll.pars[block_name].release()

    # fit
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)

    minz.plot()
    v = model(ll.pars, jac=0)

    pl.figure()
    pl.plot(dp.y, v, 'r.')

    ax = pl.figure().add_subplot(projection='3d')
    ax.errorbar(dp.x1, dp.x2, dp.y, zerr=dp.yerr, color='k', ls='', marker='.')
    ax.plot(dp.x1, dp.x2, v, 'r+')

    ax = pl.figure().add_subplot(projection='3d')
    ax.errorbar(dp.x1, dp.x2, dp.y-v, zerr=dp.yerr, color='k', ls='', marker='.')

    return minz
    
def main_snake(release=['x0', 'tmax','stretch'], error_pedestal=0.05):
    # generate dataset
    gen = lightcurves.ToyModelGenerator(nsn = 100, nbands=1, npts1=20, npts2=30, yerr=0.01, error_pedestal=error_pedestal)
    dp = gen.generate()

    # build a model and a log-likelihood
    model = lightcurves.Model(dp)
    reg = lightcurves.Regularization(block_name='theta')
    cons = lightcurves.cons(model, mu=1.E6)
    snake = lightcurves.SimpleErrorSnake_new(model)
    ll = LogLikelihood(model, variance_model=snake, reg=[reg], cons=[cons], data=dp)

    # fit initialization
    for block_name in ['x0', 'tmax', 'stretch']:
        ll.pars[block_name] = dp.sample[block_name]
        ll.pars[block_name].fix()

    for block_name in release:
        ll.pars[block_name].release()

    # fit
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)

    minz.plot()
    v = model(ll.pars, jac=0)

    pl.figure()
    pl.plot(dp.y, v, 'r.')

    ax = pl.figure().add_subplot(projection='3d')
    ax.errorbar(dp.x1, dp.x2, dp.y, zerr=dp.yerr, color='k', ls='', marker='.')
    ax.plot(dp.x1, dp.x2, v, 'r+')

    ax = pl.figure().add_subplot(projection='3d')
    ax.errorbar(dp.x1, dp.x2, dp.y-v, zerr=dp.yerr, color='k', ls='', marker='.')
    
    minz.gen = gen

    return minz
    
def main_error_pedestal(release=['x0', 'tmax'], gen_error_pedestal=0.05, gen_error_snake=None, gamma_init=0.):
    # generate dataset
    gen = lightcurves.ToyModelGenerator(nsn = 50, nbands=1, npts1=20, npts2=30, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
    print(gen.error_snake, gen.error_pedestal)
    dp = gen.generate()

    # build a model and a log-likelihood
    model = lightcurves.Model(dp)
    reg = lightcurves.Regularization(block_name='theta')
    cons = lightcurves.cons(model, mu=1.E6)
    snake = lightcurves.SimplePedestalModel(model)
    # snake = lightcurves.SimpleErrorSnake(model)
    ll = LogLikelihood(model, variance_model=snake, reg=[reg], cons=[cons], data=dp)

    # fit initialization
    for block_name in ['x0', 'tmax', 'stretch']:
        ll.pars[block_name].full[:] = dp.sample[block_name]
        ll.pars[block_name].fix()

    for block_name in release:
        ll.pars[block_name].release()

    # fit
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)

    minz.plot()
    v = model(ll.pars, jac=0)

    pl.figure()
    pl.plot(dp.y, v, 'r.')

    ax = pl.figure().add_subplot(projection='3d')
    ax.errorbar(dp.x1, dp.x2, dp.y, zerr=dp.yerr, color='k', ls='', marker='.')
    ax.plot(dp.x1, dp.x2, v, 'r+')

    ax = pl.figure().add_subplot(projection='3d')
    ax.errorbar(dp.x1, dp.x2, dp.y-v, zerr=dp.yerr, color='k', ls='', marker='.')
    
    minz.gen = gen

    return minz



def main_error_snake(nsn=50, npts1=20, npts2=15, release=['x0', 'tmax'], gen_error_pedestal=None, gen_error_snake=0.05, gamma_init=0.05):
    # generate dataset
    gen = lightcurves.ToyModelGenerator(nbands=1, nsn=nsn, npts1=npts1, npts2=npts2, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
    print(gen.error_snake, gen.error_pedestal)
    dp = gen.generate()

    # build a model and a log-likelihood
    model = lightcurves.Model(dp)
    reg = lightcurves.Regularization(block_name='theta')
    cons = lightcurves.cons(model, mu=1.E6)
    snake = lightcurves.SimpleErrorSnake(model)
    ll = LogLikelihood(model, reg=[reg], cons=[cons], data=dp)

    # fit initialization
    for block_name in ['x0', 'tmax', 'stretch']:
        ll.pars[block_name].full[:] = dp.sample[block_name]
        ll.pars[block_name].fix()

    for block_name in release:
        ll.pars[block_name].release()

    # first fit with model only
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6, dchi2=10., max_iter=10)
    minz.plot()

    # second fit, with model and error model
    ll = LogLikelihood(model, variance_model=snake, reg=[reg], cons=[cons], data=dp)
    for block_name in ['x0', 'tmax', 'stretch', 'theta']:
        ll.pars[block_name].full[:] = r['pars'][block_name].full[:]
    ll.pars['gamma'].full[:] = gamma_init
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)
    minz.plot()

    minz.gen = gen

    return minz

def main_error_snake_calib_prior(nsn=50, nbands=3, npts1=20, npts2=15, release=['x0', 'tmax', 'stretch', 'eta'], gen_error_pedestal=None, gen_error_snake=0.05, gamma_init=0.05):
    # generate dataset
    gen = lightcurves.ToyModelGenerator(nbands=nbands, nsn=nsn, npts1=npts1, npts2=npts2, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
    print(gen.error_snake, gen.error_pedestal)
    dp = gen.generate()

    # build a model and a log-likelihood
    model = lightcurves.ModelwithCalibScatter(dp)
    reg = lightcurves.Regularization(block_name='theta')
  
    cons = lightcurves.cons(model, mu=1.E6)
    snake = lightcurves.SimpleErrorSnake(model)
    V_calib = np.diag(np.full(nbands, 0.005**2))
    calib_prior = lightcurves.CalibPrior(V_calib)
    ll = LogLikelihood(model, reg=[reg], cons=[cons], priors=[calib_prior], data=dp)

    # fit initialization
    for block_name in ['x0', 'tmax', 'stretch']:
        ll.pars[block_name].full[:] = dp.sample[block_name]
        ll.pars[block_name].fix()

    for block_name in release:
        ll.pars[block_name].release()

    # first fit with model only
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6, dchi2=10., max_iter=10)
    minz.plot()

    # second fit, with model and error model
    ll = LogLikelihood(model, variance_model=snake, reg=[reg], cons=[cons], data=dp)
    for block_name in ['x0', 'tmax', 'stretch', 'theta']:
        ll.pars[block_name].full[:] = r['pars'][block_name].full[:]
    ll.pars['gamma'].full[:] = gamma_init
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)
    minz.plot()

    minz.gen = gen

    return minz

def main_lc_error_snake_calib_prior(nsn=50, nbands=3, npts1=20, npts2=15, release=['x0', 'tmax', 'stretch', 'eta', 'color'], 
                                 gen_error_pedestal=None, gen_error_snake=0.05, gamma_init=0.05):
    # generate dataset
    gen = lightcurves.ToyModelGenerator(nbands=nbands, nsn=nsn, npts1=npts1, npts2=npts2, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
    print(gen.error_snake, gen.error_pedestal)
    dp = gen.generate()

    # build a model and a log-likelihood
    model = lightcurves.LightCurve2DCalibScatter(dp)
    reg = lightcurves.Regularization(block_name='theta')
  
    cons = lightcurves.cons(model, mu=1.E6, color='True')
    snake = lightcurves.SimpleErrorSnake(model)
    V_calib = np.diag(np.full(nbands, 0.005**2))
    calib_prior = lightcurves.CalibPrior(V_calib)
    ll = LogLikelihood(model, reg=[reg], cons=[cons], priors=[calib_prior], data=dp)

    # fit initialization
    for block_name in ['x0', 'tmax', 'stretch', 'color']:
        ll.pars[block_name].full[:] = dp.sample[block_name]
        ll.pars[block_name].fix()

    for block_name in release:
        ll.pars[block_name].release()

    # first fit with model only
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6, dchi2=10., max_iter=10)
    minz.plot()

    # second fit, with model and error model
    ll = LogLikelihood(model, variance_model=snake,
                       reg=[reg], cons=[cons],
                       priors=[calib_prior],
                       data=dp)
    for block_name in ['x0', 'tmax', 'stretch', 'color','theta']:
        ll.pars[block_name].full[:] = r['pars'][block_name].full[:]
    ll.pars['gamma'].full[:] = gamma_init
    minz = Minimizer(ll)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)
    minz.plot()

    minz.gen = gen

    return minz


def plot_fit_summary(minz, nsn_to_draw=3, band_to_draw=1):
    """
    """
    ll = minz.log_likelihood
    model = ll.model
    snake = ll.variance_model
    pars = ll.pars
    dp = ll.data
    gen = minz.gen

    x0 = pars['x0'].full
    tmax = pars['tmax'].full
    stretch = pars['stretch'].full
    nsn = len(x0)

    v = model(pars, jac=0)
    var = snake(pars, jac=0)
    
    fig = pl.figure()
    gs = GridSpec(4, 3)
    idx = dp.sn < nsn_to_draw
    ax00 = fig.add_subplot(gs[0], projection='3d')
    ax00.errorbar(dp.x1[idx], dp.x2[idx], dp.y[idx], zerr=np.sqrt(dp.yerr**2 + var)[idx], color='gray', ls='', marker=',', alpha=0.25)
    _dp = gen.gen_fine_grid()
    _model = lightcurves.Model(_dp)
    if snake is not None:
        _snake = snake.__class__(_model)
        _pars = FitParameters(_model.get_struct() + _snake.get_struct())
        _model.init_pars(_pars)
        _snake.init_pars(_pars)
        _pars['gamma'].full[:] = pars['gamma'].full[:]
    else:
        _snake = None
        _pars = _model.init_pars()
    _pars['theta'].full[:] = pars['theta'].full[:]
    for isn in range(3):
        for blk in ['x0', 'tmax', 'stretch']:
            _pars[blk].full[0] = pars[blk].full[isn]
        if 'color' in _pars._struct:
            _pars['color'].full[0] = pars['color'].full[isn]
        vv = _model(_pars, jac=False)
        
        _v = _model(_pars, jac=False)
        _ev = np.sqrt(_snake(_pars, jac=False))
        
        color = pl.cm.jet(int(isn / nsn_to_draw * 256))
        #idx = dp.sn_index == isn
        idx = (dp.sn_index == isn) & (dp.band == band_to_draw)
        ax00.errorbar(dp.x1[idx], dp.x2[idx], dp.y[idx], zerr=dp.yerr[idx], color=color, ls='', marker='.', elinewidth=2)
        idx = _dp.band == band_to_draw
        ax00.plot(_dp.x1[idx], _dp.x2[idx], _v[idx], ls='-', color=color, alpha=0.25, zorder=10)
        ax00.plot(_dp.x1, _dp.x2, vv, ls='-', color=color, alpha=0.25, zorder=10)
        #ax00.fill_between(_dp.x1[idx], _dp.x2[idx], (_v-_ev)[idx], (_v+_ev)[idx], color=color, alpha=0.7)

    ax00.set_xlabel('x1')
    ax00.set_ylabel('x2')
    ax00.set_zlabel('data & model')
    
    idx = (dp.sn_index == isn) & (dp.band == band_to_draw)
    phase1 = (dp.x1[idx]-tmax[dp.sn_index][idx]) * (1. + stretch[dp.sn_index][idx])
    phase2 = (dp.x2[idx]-tmax[dp.sn_index][idx]) * (1. + stretch[dp.sn_index][idx])
    yy = dp.y[idx] / x0[dp.sn_index][idx]
    eyy = dp.yerr[idx] / np.abs(x0[dp.sn_index][idx])
    ax01 = fig.add_subplot(gs[1], projection = '3d')
    ax01.errorbar(phase1, phase2, yy, zerr=eyy, color='k', ls='', marker='.', label='fit')
    xx = np.linspace(phase1.min(), phase1.max(), 30)
    yy = np.linspace(phase2.min(), phase2.max(), 30)
    XX, YY = np.meshgrid(xx, yy)
    J = model.basis.eval(XX.ravel(), YY.ravel())
    ax01.plot(XX.ravel(), YY.ravel(), J @ pars['theta'].full, color='red', linewidth=2, zorder=100, label='model')
    ax01.set_xlabel('phase')
    ax01.set_ylabel('data & model')
    pl.legend(loc='best')

    ax02 = fig.add_subplot(gs[2])
    ax02.plot(dp.var**2, var, color='brown', marker='.', ls='')
    xx = np.linspace(0., var.max(), 100)
    ax02.plot(xx, xx, 'b--')
    ax02.set_xlabel('$V_{true}$')
    ax02.set_ylabel('$V_{reco}$')
    
    ax10 = fig.add_subplot(gs[3])
    ax10.errorbar(dp.x1[idx], dp.y[idx]-v[idx], yerr=np.sqrt(dp.yerr[idx]**2 + var[idx]), color='brown', marker=',', ls='')
    ax10.errorbar(dp.x1[idx], dp.y[idx]-v[idx], yerr=dp.yerr[idx], color='k', marker='.', ls='', elinewidth=2)
    ax10.set_xlabel('x1')
    ax10.set_ylabel('residuals[idx]')
    
    ax11 = fig.add_subplot(gs[4])
    ax11.errorbar(phase1, dp.y[idx]-v[idx], yerr=np.sqrt(dp.yerr[idx]**2 + var[idx]), color='brown', marker=',', ls='')
    ax11.errorbar(phase1, dp.y[idx]-v[idx], yerr=dp.yerr[idx]**2, color='k', marker='.', ls='')
    ax11.set_xlabel('phase1')
    ax11.set_ylabel('residuals[idx]')
    
    ax12 = fig.add_subplot(gs[5])
    ax12.plot(phase1, dp.var[idx]-var[idx], color='brown', marker='.', ls='')
    ax12.set_xlabel('phase')
    ax12.set_ylabel('$V-V_{true}$')
    
    
    ax20 = fig.add_subplot(gs[6])
    ax21 = fig.add_subplot(gs[7])
    ax22 = fig.add_subplot(gs[8])
    ax20.plot(dp.sample.x0, pars['x0'].full, 'b.')
    ax20.set_xlabel('$X_0$ [true]')
    ax20.set_ylabel('$X_0$ [reco]')
    ax21.plot(dp.sample.tmax, pars['tmax'].full, 'b.')
    ax21.set_xlabel('$x_{peak}$ [true]')
    ax21.set_ylabel('$x_{peak}$ [reco]')
    ax22.plot(dp.sample.stretch, pars['stretch'].full, 'b.')
    ax22.set_xlabel('$s$ [true]')
    ax22.set_ylabel('$s$ [reco]')
    
    free_pars = ['x0', 'tmax', 'stretch', 'gamma']
    if 'color' in pars._struct:
        ax30 = fig.add_subplot(gs[9])
        free_pars.append('color')
        ax30.plot(dp.sample.color, pars['color'].full, 'b.')
        ax30.set_xlabel('$color$ [true]')
        ax30.set_ylabel('$color$ [reco]')
    
    ax31 = fig.add_subplot(gs[10])
    ax31.plot(phase1, (dp.y[idx]-v[idx]) / np.sqrt(dp.yerr[idx]**2 + var[idx]), color='k', marker='.', ls='')
    ax31.set_xlabel('phase1')
    ax31.set_ylabel('pulls[idx]')
    
    V, err_of_interest, corr = minz.get_cov_matrix(free_pars, corr=True, plot=True)




import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas
from ....lib import bspline
from matplotlib import cm


def gaussian_2d(x, y, a, sigma):
    """
    Gaussian function *2
    """
    norm = a / (np.sqrt(2. * np.pi) * sigma)
    xx = x / sigma
    yy = y / sigma
    return norm * np.exp(-0.5 * (xx**2 + yy**2))

class ToyDataGenerator:

    def __init__(self, x0_min=-10., x0_max=10., x1_min=-10., x1_max=10., yerr=0.01, a=1., sigma=3.,
                 n0=40, n1=40, delta=2., b0_size=10, b1_size=10, order=4, error_pedestal=0.025):
        self.x0_min, self.x0_max = x0_min, x0_max
        self.x1_min, self.x1_max = x1_min, x1_max
        self.yerr = yerr
        self.error_pedestal = error_pedestal
        if self.error_pedestal is not None:
            self.sig_noise = np.sqrt(self.yerr**2 + self.error_pedestal**2)
        else:
            self.sig_noise = yerr
        self.a, self.sigma = a, sigma
        self.n0 = n0
        self.n1 = n1
        self.delta = delta
        self.basis = bspline.BSpline2D(np.linspace(x0_min-delta, x0_max+delta, b0_size), 
                                       np.linspace(x1_min-delta, x1_max+delta, b1_size), 
                                       x_order=order, y_order=order)
        #self.error_pedestal=error_pedestal

    def generate(self):
        """
        generate a realization of a dataset
        """
        x0 = np.random.uniform(self.x0_min, self.x0_max, size=self.n0)
        x1 = np.random.uniform(self.x1_min, self.x1_max, size=self.n1)
        X0, X1 = np.meshgrid(x0, x1)
        X0 = X0.ravel()
        X1 = X1.ravel()
        n = len(X0)
        self.n = n
        yerr = np.full(self.n, self.yerr)
        sig_noise = np.full(self.n, self.sig_noise)
        y = gaussian_2d(X0, X1, self.a, self.sigma)
        noise = np.random.uniform(low = -self.yerr, high = self.yerr, size = n)
        self.x0, self.x1, self.y_true, self.noise, self.yerr = x0, x1, y, noise, yerr
        self.y = self.y_true + self.noise
        return self.x0, self.x1, self.y + self.noise, self.yerr

    def get_data(self):
        if not hasattr(self, 'x0'):
            self.generate()
        X0, X1 = np.meshgrid(self.x0, self.x1)
        dd = pandas.DataFrame({'xx0': X0.ravel(),
                               'xx1': X1.ravel(),
                               'y': self.y,
                               'yerr': self.yerr,
                               'bads': np.zeros(len(self.yerr)).astype(bool),
                               })
        return dd

    #def fit(self, beta=1.E-6):
    #    """a fit of the data -- the simplest possible
    #    """
    #    X0, X1 = np.meshgrid(self.x0, self.x1)
    #    J = self.basis.eval(X0.ravel(), X1.ravel())
    #    w = 1. / self.yerr**2
    #    W = dia_matrix((w, [0]), shape=(self.n, self.n))
    #    H = J.T @ W @ J
    #    B = J.T @ W @ self.y
    #    fact = cholmod.cholesky(H, beta=1.E-6)
    #    self.theta = fact(B)

    def plot(self, color='b', marker='.'):
        """plot the data and the model
        """
        mu = self.mu
        fig = pl.figure()
        gs = GridSpec(2, 1, height_ratios=[7,2])

        ax1 = fig.add_subplot(gs[0], projection='3d')
        #ax1.errorbar(self.x, self.y, yerr=self.yerr, ls='', color=color, marker=marker, label='data', zorder=0.5)
        ax1.set_zlabel('model & data')
        ax1.set_xlabel('x0')
        ax1.set_ylabel('x1')
        ax2 = fig.add_subplot(gs[1], sharex=ax1)
        ax2.set_xlabel('x')
        ax2.set_ylabel('residuals')
        
        X0, X1 = np.meshgrid(self.x0, self.x1)
        ax1.errorbar(X0.ravel(), X1.ravel(), self.y, zerr = self.yerr, ecolor=('black',0.1), ls='', marker='.', color='blue')
        if self.mu ==0.:
            ax1.set_title('µ0 = ' +  str(self.mu_0) + ' , µ1 = ' + str(self.mu_1))
        else : 
            ax1.set_title('µ max = ' + str(self.mu))
            

        # pl.grid(ls=':')
        # pl.plot(self.basis.grid, np.zeros(len(self.basis.grid)), 'k|')
        
        resid = []
        errs = []
        if hasattr(self, 'theta'):
            #xx = self.basis._grid.copy()
            xx = self.basis.bx.grid
            yy = self.basis.by.grid
            XX, YY = np.meshgrid(xx, yy)
            JJ = self.basis.eval(XX.ravel(), YY.ravel())
            #ax1.plot(xx, yy, JJ@self.theta, ls='', color='red', marker='.', alpha=1., label='nodes', zorder=10., linewidth=2)
            fitting = JJ@self.theta
            fitting = np.reshape(fitting, (self.b0, self.b1))
            ax1.plot_surface(XX, YY, fitting, cmap=cm.coolwarm, label='nodes')

            #xx = np.linspace(self.x0_min-self.delta, self.x0_max+self.delta, 200)
            #JJ = self.basis.eval(xx)
            #ax1.plot(xx, JJ@self.theta, 'r-', label='model', zorder=2.)

            JJ = self.basis.eval(X0.ravel(), X1.ravel())
            x = np.linspace(self.x0_min, self.x0_max, len(X0.ravel()))
            ax2.errorbar(x, self.y-JJ @ self.theta, yerr=self.yerr, ls='', marker='.', color='k')
            
            residuals_data = self.y-JJ @ self.theta
            pl.figure()
            pl.hexbin(X0.ravel(), X1.ravel(), residuals_data/self.yerr, gridsize=15)
            pl.colorbar()
            pl.title('Weighted Residuals')
            
        pl.legend(loc='best')

def plot_model_and_data(model, pars, data, title='', reg_lambda=None):
    """
    """
    fig = plt.figure()
    #gs = GridSpec(3, 1, height_ratios=[4,2,2])
    gs = GridSpec(2, 1, height_ratios=[7,2])
    ax1 = fig.add_subplot(gs[0], projection='3d')
        
    xx0 = np.linspace(model.basis.bx.grid.min(), model.basis.bx.grid.max(), 200)
    xx1 = np.linspace(model.basis.by.grid.min(), model.basis.by.grid.max(), 200)    
    XX0, XX1 = np.meshgrid(xx0, xx1)    
    J = model.basis.eval(XX0.ravel(), XX1.ravel())    
    YY = J@pars['theta'].full
    YY = np.reshape(YY, (200,200))
    ax1.plot_surface(XX0, XX1, YY, cmap=cm.coolwarm)  
    
    ax1.errorbar(data.xx0, data.xx1, data.y, zerr=data.yerr, ecolor=('black',0.1), ls='', marker='.', color='blue')  
    #plt.setp(ax1.get_xticklabels(), visible=False)
    ax1.set_zlabel('model \& data')
    
    ax2 = fig.add_subplot(gs[1], sharex=ax1)
    res = data.y-model(pars)
    ax2.errorbar(data.xx0, res, yerr=data.yerr, ls='', marker='.')
    ax2.set_xlabel('x')
    ax2.set_ylabel('residuals')
    #plt.setp(ax2.get_xticklabels(), visible=False)

    #ax3 = fig.add_subplot(gs[2], sharex=ax1)
    #pars = pars.copy()
    #xx = np.linspace(model.basis.grid.min(), model.basis.grid.max(), 500)
    #J = model.basis.eval(xx)
    #for i in range(len(pars['theta'].full)):
    #    pars['theta'].full[:] = 0.
    #    pars['theta'].full[i] = 3 * res.std()
    #    if reg_lambda is not None:
    #        color = plt.cm.jet(int(reg_lambda[i]/reg_lambda.max() * 256))
    #        # color = 'r' if reg_lambda[i]>0. else 'b'
    #    else:
    #        color = plt.cm.jet(int(i * 256/len(pars['theta'].full)))
    #    ax3.plot(xx, J@pars['theta'].full, ls='-', color=color)
    #ax3.set_xlabel('x')

    #plt.subplots_adjust(hspace=0.05)
    #fig.suptitle(title)



#!/usr/bin/env python3
import os
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)
import pandas
import glob
import nacl
from nacl.dataset import TrainingDataset
from lemaitre import bandpasses

bandpass_convert = {
    'MEGACAMPSF::g': 'megacam6::g',
    'MEGACAMPSF::r': 'megacam6::r',
    'MEGACAMPSF::i2': 'megacam6::i2',
    'MEGACAMPSF::z': 'megacam6::z',
    'MEGACAMPSF::i': 'megacampsf::i',

    'megacam6::g': 'megacam6::g',
    'megacam6::r': 'megacam6::r',
    'megacam6::i2': 'megacam6::i2',
    'megacam6::z': 'megacam6::z',
    'megacam6::i': 'megacampsf::i',

    'ztfg': 'ztf::g',
    'ztfr': 'ztf::r',
    'ztfi': 'ztf::I',

    'ztf::g': 'ztf::g',
    'ztf::r': 'ztf::r',
    'ztf::I': 'ztf::I',
}


def load_sample(path, sample='SNLS', init_sn_loc=True, extension="parquet.gzip"):
    """
    load and sanitize the SNLS data

    Examples
    --------
    >>> sne, lc, sp = load_sample(path="data/test_datasets/", sample="SNLS")
    """
    if "parquet" in extension:
        logging.info(f'Loading {sample} from\n'+"\n".join(glob.glob(os.path.join(path, f"*_{sample}_blind.parquet.gzip"))))
        sne = pandas.read_parquet(os.path.join(path, f'data_{sample}_blind.parquet.gzip'))
        lc = pandas.read_parquet(os.path.join(path, f'lc_{sample}_blind.parquet.gzip'))
        sp = pandas.read_parquet(os.path.join(path, f'sp_{sample}_blind.parquet.gzip'))
    elif "csv" in extension:
        logging.info(f'Loading {sample} from\n'+"\n".join(glob.glob(os.path.join(path, f"*_{sample}_blind.csv"))))
        sne = pandas.read_csv(os.path.join(path, f'data_{sample}_blind.csv'))
        lc = pandas.read_csv(os.path.join(path, f'lc_{sample}_blind.csv'))
        sp = pandas.read_csv(os.path.join(path, f'sp_{sample}_blind.csv'))
    else:
        raise ValueError(f"Unknown extension: {extension}. Must be either 'parquet' or 'csv'")

    # purge all the SN which could not be fitted with sncosmo
    idx = sne['x0'] != 999.9
    logging.info(f'removed {(~idx).sum()} SNe which could not be fitted with sncosmo')
    sne = sne[idx]

    # purge all the LC data with no counterpart in the SN index
    idx = lc.name.isin(sne.name)
    lc = lc[idx]
    logging.info(f'removed {(~idx).sum()} LC data points with no entry in the SN index')
    lc.sn = lc.name

    # purge all the LC data taken with bandpass megacampsf::i
    idx = (lc.band == 'MEGACAMPSF::i') | (lc.band == 'megacampsf::i') | (lc.band == 'megacam6::i')
    lc = lc[~idx]
    logging.info(f'removed {(idx).sum()} LC data points taken in band megacampsf::i')

    # get rid of AB_B12 magsys
    # we assume that the light curves will be recalibrated
    #    idx = lc.magsys == 'AB_B12'
    #    lc.loc[idx, 'magsys'] = 'AB'
    #    logging.info(f'hacked the magsys of {idx.sum()} measurements from AB_B12 to AB')

    # purge all the spectra with no counterpart in the SN index
    idx = sp.name.isin(sne.name)
    sp = sp[idx]
    logging.info(f'removed {(~idx).sum()} SP data points with no entry in the SN index ')
    sp.sn = sp.name

    # correct the valid field
    sp.loc[sp.valid.isna(), 'valid'] = 1
    sp['valid'] = sp.valid.astype(int)

    # add a i_basis field
    sp['i_basis'] = 0

    # we use names to identify the SNe
    sne.sn = sne.name
    lc.sn = lc.name
    sp.sn = sp.name

    # update the bandpass names
    lc['band'] = [bandpass_convert[b] for b in lc.band]
    if 'survey' not in sne.columns:
        lc['survey'] = sample

    if init_sn_loc:
        lc['x'] = 0.
        lc['y'] = 0.
        lc['ccd'] = 0
        lc['amp'] = 0

    return sne, lc, sp


def dataset_release(path):
    """
    Build a TrainingDataset from an existing path

    :param path: str
    :return: nacl.dataset.TrainingDataset

    Examples
    --------
    >>> tds = dataset_release("../../../data/lemaitre_light/")
    >>> assert isinstance(tds, nacl.dataset.TrainingDataset)
    """
    sne_ztf, lc_ztf, sp_ztf = load_sample(path, 'ztf')
    lc_ztf['x'] = 1000.
    lc_ztf['y'] = 1000.
    lc_ztf['sensor_id'] = 12

    sne_SNLS, lc_SNLS, sp_SNLS = load_sample(path, 'SNLS')
    lc_SNLS['x'] = 1000.
    lc_SNLS['y'] = 1000.
    lc_SNLS['sensor_id'] = 12
    lc_offset = lc_ztf.lc.max() + 1
    lc_SNLS['lc'] = lc_SNLS['lc'] + lc_offset
    spec_offset = sp_ztf.spec.max() + 1
    sp_SNLS['spec'] = sp_SNLS['spec'] + spec_offset

    flib = bandpasses.get_filterlib(rebuild=False)
    tds = TrainingDataset(pandas.concat([sne_ztf, sne_SNLS]),
                          pandas.concat([lc_ztf, lc_SNLS]),
                          pandas.concat([sp_ztf, sp_SNLS]),
                          filterlib=flib)
    return tds
